import esprima

# Define visitor functions
def visit_node(node, parent=None):
    if node is None:
        return
        
    node_type = node.type
    

    # Handle export declarations
    if node_type == 'ExportNamedDeclaration' or node_type == 'ExportDefaultDeclaration':
        # The actual declaration is nested inside the export
        if hasattr(node, 'declaration') and node.declaration:
            return visit_node(node.declaration, node)
        return None

    if node_type == 'ClassDeclaration':
        class_name = node.id.name if hasattr(node, 'id') and node.id else 'anonymous'
        return ["class", class_name]
    
    elif node_type == 'FunctionDeclaration':
        func_name = node.id.name if hasattr(node, 'id') and node.id else 'anonymous'
        return ["function", func_name]

    elif node_type == 'VariableDeclaration':
        declarations = node.declarations
        if declarations:
            declaration = declarations[0]
            var_name = declaration.id.name if hasattr(declaration, 'id') and declaration.id else 'unknown'
            init_node = declaration.init

            init_type = init_node.type if init_node else ''
            
            return [init_type, var_name]
        
    elif node_type == 'IfStatement':
        return "ifStatement", "ifStatement"
    
    return None

def removeDefaultKeywordFromChunk(chunk):
    """
    Replace the 'export default' keyword with 'const' so a parser can 
    successfully process and extract variable name.
    
    Args:
        chunk (str): code to process

    Returns:
        str: processed chunk

    """

    chunk_stripped = chunk.strip().rstrip(';')

    if not chunk_stripped.startswith("export default "):
        return chunk
    
    # Remove "export default " prefix and add "const " 
    rest = chunk_stripped[len("export default "):]

    return f"const {rest}"


def get_variable_name_and_type(chunk):    
    # Parse the JavaScript code

    if chunk.strip().startswith("/") or "=" not in chunk:
        return None 
    
    if chunk.strip().startswith("export default "):
        chunk = removeDefaultKeywordFromChunk(chunk)
    
    try:
        # Use parseModule for ES6 module syntax (export/import)
        ast = esprima.parseModule(chunk)
        info = visit_node(ast.body[0])
        return info
    except esprima.error_handler.Error as e:
        raise ValueError(f"Parse error: {e} could not parse code : {chunk}")
        # return None
    except IndexError:
        raise ValueError("error processing chunk : ",chunk)


import os

from speedbuild.auth.cli_utils import get_config

provider_to_api_key_name = {
    "openai":"OPENAI_API_KEY",
    "anthropic":"ANTHROPIC_API_KEY",
    "google":"GEMINI_API_KEY",
    "gemini":"GEMINI_API_KEY"
}

def getProviderAPIKey(provider):
    if provider == None:
        return None
    return os.environ.get(provider_to_api_key_name[provider],None)

def getLLMConfig():
    model_provider = os.environ.get("speedbuild_model_provider",None)
    model_name = os.environ.get("speedbuild_model_name",None)
    provider_key = getProviderAPIKey(model_provider)

    if model_provider == None:
        raise ValueError("Model provider has not been specified")
    
    if model_name == None:
        raise ValueError("Model name has not been specified")
    
    if provider_key == None:
        raise ValueError(f"API key for model provider {model_provider} has not been set")
    
    return [model_provider,model_name]


def setSpeedbuildConfig():
    config = get_config()
    config_keys = config.keys()

    if "default_model" in config_keys:
        model_name = os.environ.get("speedbuild_model_name",None)
        model_name_in_config = config['default_model']
        
        if model_name_in_config != model_name:
            os.environ['speedbuild_model_name'] = model_name_in_config

    if "model_provider" in config_keys:
        model_provider = os.environ.get("speedbuild_model_provider",None)

        if config['model_provider'] != model_provider:
            os.environ["speedbuild_model_provider"] = config["model_provider"]

    if "llm_keys" in config_keys and "model_provider" in config_keys:
        model_provider = config['model_provider']
        key_config = config["llm_keys"]

        provider_key = getProviderAPIKey(model_provider)
        key_name = provider_to_api_key_name[model_provider]

        if provider_key != key_config[model_provider]:
            os.environ[key_name] = key_config[model_provider]


import sys

django_workflow = """
    ## Workflow Strategy

    ### 1. Understand Request
    - Analyze the customization goal and scope
    - Identify affected Django components (models, views, templates, forms, serializers, middleware)
    - Determine file processing order and dependencies

    ### 2. Gather Context
    - Read relevant files to understand current implementation
    - Check Django settings, URLs configuration, and app structure
    - Identify potential breaking changes or dependencies
    - Review migration files if model changes are involved

    ### 3. Execute Changes
    - **Process dependencies first** before main feature modifications
    - Make **one change at a time** and verify impact
    - Follow Django patterns and conventions
    - Add meaningful docstrings and comments for complex changes

    ### 4. Validate Results
    - Run Django management commands to verify functionality
    - Check for broken imports or syntax errors
    - Ensure migrations are properly generated if needed
    - Ensure no regression in existing features

    IMPORTANT NOTES : 
    1. Always read_file() before making changes
    2. Use break_chunk() to isolate the exact code you need to modify
    3. Use update_file() to make the change
    4. Verify your changes make syntactic and logical sense

    General Rules:

    - Always inspect files using read_file when needed.
    - Always follow clean, production-grade Django & Python best practices.
    - Perform edits ONLY using update_file.
    - Use break_chunk to isolate specific code before editing.
    - If an action fails after multiple attempts:
        Respond with a JSON explaining that the action was not successful and include possible reasons.
    - Never repeat failed attempts unnecessarily.
"""

express_workflow = """
    ## Workflow Strategy

    ### 1. Understand Request
    - Analyze the customization goal and scope
    - Identify affected Express.js components (routes, middleware, controllers, models, services)
    - Determine file processing order and dependencies

    ### 2. Gather Context
    - Read relevant files to understand current implementation
    - Check configurations, dependencies, and environment setup
    - Identify potential breaking changes or dependencies

    ### 3. Execute Changes
    - **Process dependencies first** before main feature modifications
    - Follow Express Js patterns and conventions
    - Add meaningful docstrings and comments for complex changes

    ### 4. Validate Results
    - Run tests or commands to verify functionality
    - Check for broken imports or syntax errors
    - Ensure no regression in existing features

    IMPORTANT NOTES : 
    1. Always read_file() before making changes
    2. Use break_chunk() to isolate the exact code you need to modify
    3. Use update_file() to make the change
    4. Verify your changes make syntactic and logical sense

    General Rules:

    - Always inspect files using read_file when needed.
    - Always follow clean, production-grade NodeJs & Express Js best practices.
    - Perform edits ONLY using update_file.
    - Use break_chunk to isolate specific code before editing.
    - If an action fails after multiple attempts:
        Respond with a JSON explaining that the action was not successful and include possible reasons.
    - Never repeat failed attempts unnecessarily.

"""

tools_and_response_format = """
    Available Tools

    1. read_file
    - Reads and returns the content of a file.
    - Arguments:
        file_name (str): Absolute path of the file to read.

    2. update_file
    - Allows you to add, replace, or remove chunks of code in a file.
    - Requires:
        chunk_name (str): Name of the anchor chunk.
    - Behavior:
        add → write new code *before* the anchor chunk
        replace → new code *replaces* the anchor chunk
        remove → anchor chunk is *deleted*
    - The anchor chunk is mandatory for all write operations.
    - Never modify code that does not belong to a named chunk.

    3. break_chunk
    - Splits an existing chunk into smaller named sub-chunks.
    - Use this when you must modify part of a chunk.
    - You may call break_chunk repeatedly until the exact code you need to modify has its own named chunk.
    - Never perform update_file operations on unnamed code.

    Shell Commands:
    - You CANNOT execute shell commands directly
    - To request command execution or package installation, respond with EXACTLY this JSON format:
    {
        "type": "shell_command",
        "command": <Command to run>,
        "message": <short single active non-request sentence description of Command action. e.g "starting the developtment server">
    }
    - STOP and WAIT for the execution result before continuing

    Regular Response:
    - for regular non command execution use the json format:
    {
        "message": <normal response>
    }

    ### Error Responses
    If something fails or you cannot resolve issue, respond:

    {
        "type": "error"
        "message": "<error message>"
    }
"""

platform = """
---
    NOTE : The application is running on """


if sys.platform == "win32":
    platform += "Windows  !!!"
else:
    platform += "a Unix-based os !!!"


platform += "\n---"


def get_system_prompt(framework,files=None,debugging=False):

    if framework == "django":
        framework = "Django"
        workflow_instruction = django_workflow
    else:
        framework = "Express JS"
        workflow_instruction = express_workflow

    return f"""
    You are a **smart AI agent** specialized in **{framework}** and experienced in **production-grade {framework} code modification**. 
    You are part of SpeedBuild, a service that helps developers extract, customize, and reuse {framework} features using plain English commands.

    ## Core Mission
    Extract, Debug and customize {framework} features while maintaining production reliability, performance, and {framework} best practices.

    {
     "Debug Instruction : You are given error logs, stack traces, or failing tests and must identify and fix bugs." if debugging else ""   
    }

    ### Environment Variables Rule
    - If a feature requires environment variables, you must **only inform the user** which variables they need.
    - You must **not attempt to fix or debug errors caused by missing environment variables**.
    - Do not inspect environment-dependent conditions in the system.

    IMPORTANT:
    - You must ALWAYS respond with valid JSON ONLY.
    - Do not output explanations, markdown, or any text outside JSON.
    - Your output will be parsed by a computer program.
    - ignore environmental variables and just inform the user of the env variables needed, do not process for environmental variable related issues

    {tools_and_response_format.strip()}

    {"Project Files:" if files is not None else ""}

    {files if files is not None else ""}

    {"Note: These are the ONLY files you can modify. Always use read_file() to inspect contents before editing." if files is not None else ""}

    {workflow_instruction.strip()}

    {platform}

    """ 


if __name__ == "__main__":
    print(get_system_prompt("django"))


import os
from pathlib import Path


def getCurrentDjangoFiles(path):
    python_files = [file for file in os.listdir(path) if file.endswith(".py")]
    return python_files


def findFilePath(path,filename):
    "Get files in a directory"
    file_paths = []
    for root, _, files in os.walk(path):
        for file in files:
            # Get the relative path to the base directory
            if filename in file:
                relative_path = os.path.relpath(os.path.join(root, file), path)
                file_paths.append(relative_path)

    return file_paths


def getAbsolutePath(relative_path):
    relative_path = Path(relative_path)
    absolute_path = relative_path.resolve()

    return str(absolute_path)

def get_template_output_path(feature_name):
    user_dir = str(Path.home())
    user_dir = os.path.join(user_dir,".sb_zip")

    if not os.path.exists(user_dir):
        os.makedirs(user_dir)

    output_zip = os.path.join(user_dir,f"speed_build_{feature_name}")  # No .zip extension needed

    return output_zip


import os
from pathlib import Path

from .js_var_names import get_variable_name_and_type


output_marker = "### PARSER OUTPUT ###"
varIdentifies = ["const ","let ","var "]
special_case = [".",",","else","||","&&"]

def getLineValue(line):
    """
    Split a string at the first '=' and return the two parts.
    Searches the given string for the first occurrence of the '=' character.
    If found, returns a two-element list [left, right] where `left` contains
    the substring up to and including that first '=', and `right` contains
    the remainder of the string after the '='. If no '=' is present, returns
    None.
    Parameters
    ----------
    line : str
        The input string to examine. The function expects a string; passing
        other types may raise a TypeError.
    Returns
    -------
    list[str, str] | None
        A list with two strings [left_including_equals, right_after_equals]
        when '=' is found; otherwise None.
    Examples
    --------
    getLineValue("key=value")  -> ["key=", "value"]
    getLineValue("a=b=c")      -> ["a=", "b=c"] TODO : implement multi = sign
    getLineValue("novalue")    -> None
    """
    if "=" in line:
        pos = line.index("=") + 1
        return [line[:pos],line[pos:]]
    
    return None

class JsTxParser:
    def __init__(self):
        self.opening = [ "[","(","{","/*"]
        self.closing = ["]",")","}","*/"]
        self.cache = None

    def checkIfChunkIsContinuation(self,chunk):
        chunk = chunk.strip()
        for item in special_case:
            if chunk.startswith(item):
                return True
        return False

    def get_and_set_variable_name(self,chunk):
        return get_variable_name_and_type(chunk)
    
    def getFileInfoFromCache(self,filename):
        if self.cache is not None and filename in self.cache.keys():
            return self.cache[filename]
        return None

    def storeFileInfoInCache(self,filename, fileData):
        # TODO : find out; is this optimized
        cache = self.cache if self.cache is not None else {}
        cache[filename] = fileData

        self.cache = cache

    def clearCache(self):
        self.cache = None

    async def parse_code(self,filename, ignore_import=False, ignore_chunk_name=False,raw_code=False):

        # check cache if file have been processed before
        cache_file_data = self.getFileInfoFromCache(filename)

        if cache_file_data is not None:
            return cache_file_data

        stack = []
        chunks = []
        memory = []
        imports = []

        variableToChunkMapping = {}

        if raw_code:
            code = filename
        else:
            with open(filename,"r") as file:
                code = file.read()

        lines = code.split("\n")
        for line in lines:
            is_import = False
            memory.append(line)

            # handle multi line comments
            if line.startswith("/*") and line.endswith("*/"):
                pass # skip edge case where multi line comment are on 1 line. E.g : - /*Hello my Guy We land shjgsgs shgshgs*/ -

            elif line.startswith("/*"):
                # print("multi line comment opening found")
                stack.append("/*")

            elif line.endswith("*/"):
                # print("multi line comment closing found")
                if len(stack) > 0:
                    last = stack[-1]
                    index = self.opening.index(last)

                    if index == self.closing.index("*/"):
                        stack.pop(-1)

                else:
                    raise ValueError("Your code has an error and its not syntactic correct")
                
            else:
                for char in line:
                    if char in self.opening:
                        stack.append(char)
                    elif char in self.closing:
                        if len(stack) > 0:
                            last = stack[-1]
                            index = self.opening.index(last)

                            if index == self.closing.index(char):
                                stack.pop(-1)

                        else:
                            raise ValueError("Your code has an error and its not syntactic correct")

            if len(stack) == 0:
                newCode = "\n".join(memory)
                if len(newCode.strip()) > 0:
                    # TODO check if newCode start with special characters like ||, else, else if ....
                    if (newCode.strip()[0] in self.opening or self.checkIfChunkIsContinuation(newCode)) and len(chunks) > 0:
                        last_chunk = chunks.pop(-1)
                        newCode = f"{last_chunk}\n{newCode}"

                    if not ignore_import and newCode.strip().startswith("import "):
                        imports.append(newCode)
                    else:
                        chunks.append(newCode)

                        if not newCode.startswith("/") and not ignore_chunk_name:

                            if newCode.strip().startswith("exports."):
                                chunk_to_process = "const "+newCode.strip()[8:]
                            else:
                                chunk_to_process = newCode

                            # print(chunk_to_process,"\n\n")
                            try:
                                chunk_info = self.get_and_set_variable_name(chunk_to_process)
                                # print(chunk_info, " - chunk info")
                                if chunk_info is not None:
                                    if isinstance(chunk_info,list):
                                        # print("info ",chunk_info[1])
                                        variableToChunkMapping[chunk_info[1]] = newCode
                                        # print("mapping ", variableToChunkMapping)
                                else:
                                    # print("chunk evaluated to none ",chunk_to_process)
                                    pass
                            except ValueError as e:
                                pass
                                # print(e)

                memory = []

        if len(stack) > 0:
            raise ValueError("Your code has an error and its not syntactic correct")
            
        import_deps = {}
        require_synthax_import = []
        new_chunks = []

        for chunk in chunks:
            require_check = "require(" in chunk

            # convert require to import syntax
            if require_check:
                chunk = convertRequireToImport(chunk)
                imports.append(chunk)
                require_synthax_import.append(chunk)

                # TODO: remove chunk from chunks
                # chunks.remove(chunk)
            else:
                new_chunks.append(chunk)

        chunks = new_chunks

        for chunk in imports:

            if " from " in chunk:
                deps, src = chunk.split(" from ")

                namedImports = []

                if "{" in deps and "}" in deps:
                    namedImports = deps.split("}")[0].split("{")[-1].split(",")


                deps = deps.replace("{","").replace("}",'').replace("import","").strip().split(",")
                src = src.replace("'","").replace('"',"")

                if src.endswith(";"):
                    src = src[:len(src)-1]

                src_deps = []

                if src in import_deps.keys():
                    src_deps = import_deps[src]

                
                for word in deps:
                    if " as " in word:
                        word, alias = word.split(" as ")
                        src_deps.append({"dep":word.strip(),"alias":alias.strip(),"standalone": word not in namedImports, "use-require":chunk in require_synthax_import})
                    
                    elif word.strip() == "*":

                        # Example paths
                        path1 = Path(os.path.dirname(filename))
                        path2 = Path(f"{src}.js")

                        merged = path1 / path2

                        _, chunks, varMappings = self.parse_code(merged,True)
        
                        src_deps[-1]['ext_deps'] = list(varMappings.keys())

                        for ext_dep in list(varMappings.keys()):
                            src_deps.append({"dep":ext_dep.strip(),"alias":None,"standalone":True,"use-require":chunk in require_synthax_import})

                    else:
                        src_deps.append({"dep":word.strip(),"alias":None,"standalone": word not in namedImports, "use-require":chunk in require_synthax_import})

                    import_deps[src] = src_deps
            else:
                pass
        
        returnObject = [imports,chunks,variableToChunkMapping,import_deps]

        # Store file data in cache
        self.storeFileInfoInCache(filename,returnObject)

        return returnObject 

def removeJsIndentifier(line):
    for indentify in varIdentifies:
        if line.startswith(indentify):
            return line.replace(indentify,"").strip()
    
    return None


def convertRequireToImport(line):
    lineData = getLineValue(line)
    if lineData is not None:
        beforeEqualSign, AfterEqualSign = lineData

        # get path
        end = -1
        start = 0

        while AfterEqualSign[start] != "(":
            start += 1

        while AfterEqualSign[end] != ")":
            end -= 1

        path = AfterEqualSign[start+1:end]
        path = path.replace("'","").replace('"',"")

        # get dependencies
        deps = removeJsIndentifier(beforeEqualSign[:-1])

        if " :" in deps:
            deps = deps.replace(" :","as")
        elif ":" in deps:
            deps = deps.replace(":"," as")

        return f'import {deps} from "{path}"'




import os
import sys
from pathlib import Path

from speedbuild.experimentation.v2.agent import SpeedBuildAgentV2
from speedbuild.utils.cleanup.template.clear_folder import clear_folder 
from speedbuild.utils.exec.runCommand import PythonExecutor 
from speedbuild.agents.debugCustomizer.agent import SpeedBuildWorkflow 
from speedbuild.utils.exec.sb_exec_utils import create_virtual_environment, install_packages_in_venv 

from ..deployment.deploy_features import implementFeature

class SpeedBuildTestAgent():
    def __init__(self):
        self.home = str(Path.home())
        self.environment_sb_path = os.path.join(self.home,".sb","environment")
        self.executor = PythonExecutor()
        venv_path = os.path.join(self.home,".sb","environment","django","venv")

        if sys.platform == "win32":
            python_path =  os.path.join(venv_path, "Scripts", "python.exe")
        else:
            python_path = os.path.join(venv_path, "bin", "python")

        # Modify environment to use the venv
        self.env = os.environ.copy()
        self.env['VIRTUAL_ENV'] = venv_path
        
        if sys.platform == "win32":
            self.env['PATH'] = f"{os.path.join(venv_path, 'Scripts')};{self.env['PATH']}"
        else:
            self.env['PATH'] = f"{os.path.join(venv_path, 'bin')}:{self.env['PATH']}"


    def installAndCreateDjangoProject(self,venv):
        res = create_virtual_environment(venv)
        installation_successful = install_packages_in_venv(os.path.join(venv,"venv"),["django"])
        # print("## virtual environment is ",self.env)
        if installation_successful:

            self.venv_path = os.path.join(venv,"venv")
            # print("### here")
            # create django project
            res = self.executor.runCommand(
                ["django-admin","startproject","speedbuild_project"],
                True,
                env=self.env,
                cwd=os.path.join(self.home,".sb","environment","django")
            )

            # TODO : scheck response to see if it was succesful

            # create django app
            res = self.executor.runCommand(
                ["python","manage.py","startapp","speedbuild_app"],
                True,
                env=self.env,
                cwd=os.path.join(self.home,".sb","environment","django","speedbuild_project")
            )

            # print(res)

            # add urls.py file
            with open(os.path.join(self.home,".sb","environment","django","speedbuild_project","speedbuild_app","urls.py"),"w") as app_url:
                app_url.write("""from django.urls import path\n\n# speedbuild_app\nurlpatterns = [\n\n]""".strip())

            # shutil.copy("/home/attah/Documents/sb_final/speedbuild/exec/app_urls.py", "/home/attah/.sb/environment/django/speedbuild_project/speedbuild_app/urls.py")

            # include new urls file in main urls file
            with open(os.path.join(self.home,".sb","environment","django","speedbuild_project","speedbuild_project","urls.py"),"w") as main_url:
                main_url.write("""from django.contrib import admin\nfrom django.urls import path, include\n\nurlpatterns = [\n\tpath('admin/', admin.site.urls),\n\tpath('speedbuild/',include("speedbuild_app.urls"))\n]""".strip())
            
            # shutil.copy("/home/attah/Documents/sb_final/speedbuild/exec/main_urls.py", "/home/attah/.sb/environment/django/speedbuild_project/speedbuild_project/urls.py")


    def createTestEnvironment(self,framework):
        environment_path = os.path.join(self.environment_sb_path,framework)

        # clear out existing project
        clear_folder(environment_path)

        if not os.path.exists(environment_path):
            os.makedirs(environment_path,exist_ok=True)

        action = self.installAndCreateDjangoProject(environment_path)

        return os.path.join(environment_path,"speedbuild_project")


    def testAndFixErrors(self):
        # run server -- check for error / test feature, troubleshoot and fix issues
        pass


    def deployFeatureInTestEnvironment(self,template_path,project_path):
        """Call speedbuild function to deploy code in test environment and also add speedbuild_app to installed apps"""

        implementFeature(template_path,project_path,"speedbuild_app",None,True,self.venv_path)


    def migrateTestAndFixErrors(self,max_debugger_count=3):
        commands = [
            "python manage.py makemigrations",
            "python manage.py migrate",
            "python manage.py runserver"
        ]

        speedbuild_agent = SpeedBuildAgentV2("django",True)

        # if prompt == None: # normal debugging
        
        # Create workflow for debugging a specific command
        llm_agent = SpeedBuildWorkflow("django")
        llm_agent.create_debugger_workflow() # remove this
            
        # # Run the initial failing command to get the first error
        # llm_agent.run(context)

        ran_debug = False
        
        for command in commands:
           # Get initial error and command output
            no_errors,stdout,stderr = llm_agent.run_shell_command(command)

            if not no_errors:
                ran_debug = True
                # llm_agent.last_error = stderr 
                # llm_agent.last_command_output = stdout
                # llm_agent.original_command = command

                command_info = f"""
                command : {command}
                --------------------------------

                stdout : {stdout}
                --------------------------------

                error : {stderr}
                """

                speedbuild_agent.run(command_info)
        
        return ran_debug

    def getPackagesinVenv(self):
        collected_packages = set()
        res,err = self.executor.runCommand(
            ["pip","list"],
            False,
            env=self.env,
            cwd=os.path.join(self.home,".sb","environment","django")
        )
        
        for i in res[2:]:
            pac = [x.strip() for x in i.split(" ") if len(x)>0]
            if(len(pac)==2):
                collected_packages.add(f"{pac[0]}=={pac[1]}")

        return collected_packages

    def TestFeature(self,template_path):
        project_path = self.createTestEnvironment("django")
 
        # deploy template [add speedbuild_app to INSTALLED_APPS settings file]
        self.deployFeatureInTestEnvironment(template_path,project_path)

        packages = self.getPackagesinVenv()

        # raise ValueError("error")

        # fix error
        ran_debug = self.migrateTestAndFixErrors()

        return [ran_debug,packages]


    def customize_feature_code(self,prompt,files,root_path):

        speedbuild_agent = SpeedBuildAgentV2("django")

        # llm_agent = SpeedBuildWorkflow("django")
        # if prompt == None: # normal debugging
        
        # Create workflow for debugging a specific command
        # llm_agent.create_debugger_workflow("customization")

        # llm_agent.last_command_output = f"""
        # customization prompt : **{prompt}**
        # project root path : **{root_path}***
        # project files : {files}
        # """

        # llm_agent.run()
        speedbuild_agent.run(prompt)

        # if llm_agent.success:
        #     print("Customization successful")
        #     return True
        # else:
        #     print("workflow ended without resolution")
        #     return False



import os
from pathlib import Path

from speedbuild.utils.exec.sb_exec_utils import install_packages_in_venv
from speedbuild.utils.markdown.yaml import read_yaml 
from speedbuild.agents.break_query.break_prompt import splitPrompt 
from speedbuild.frameworks.django.utils.venv_utils import get_activated_venv 
from speedbuild.frameworks.django.extraction.test_agent import SpeedBuildTestAgent 
from speedbuild.frameworks.django.deployment.move_django_to_target import moveFilesToTargetProject 

def getTemplateFileNames(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            file_paths.append(absolute_path)
    return file_paths

def getProjectFiles(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root or "__pycache__" in root or "migrations" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            file_name = absolute_path.replace(path,"").lstrip("/")
            if os.path.basename(file_name) not in ["__init__.py","apps.py",".speedbuild","manage.py"] and file_name.endswith("sqlite3")==False:
                file_paths.append(file_name)
    return file_paths


def testCustomizeAndDeployDjangoFeature(feature_template,target_destination,target_app_name,customize=False):
    files = getProjectFiles(target_destination)
    target_project_name = None

    for file in files:
        if os.path.basename(file) == "settings.py":
            target_project_name = os.path.dirname(file)
    
    if target_project_name == None:
        raise ValueError("Could not detect project name, could not locate the settings.py file")
    
    
    test_environment = os.path.join(str(Path.home()),".sb","environment","django","speedbuild_project") 
    djangoAgent = SpeedBuildTestAgent()


    djangoAgent.TestFeature(feature_template) #debug feature

    # ask for customization prompt
    if customize:
        prompt = input("Enter customization prompt, be as detailed as possible : \n")

        # TODO : break prompt into simple steps
        customize_prompts = splitPrompt(prompt)

        for prompt in customize_prompts:
            print("Processing",prompt)
            files = getProjectFiles(test_environment) 
            status = djangoAgent.customize_feature_code(prompt,files,test_environment)

        # TODO : if could not customize break out 

    # move feature code to target destination
    moveFilesToTargetProject(
        test_environment,
        target_destination,
        target_project_name,
        target_app_name
    )

    # install packages in target destination
    # print("Installing Feature dependencies")

    config_name = feature_template.replace("speed_build_","")
    template_root = os.path.join(str(Path.home()),".sb","sb_extracted") #os.path.join(str(Path.home()),".sb_zip")
    venv = get_activated_venv()

    yaml_path = os.path.join(template_root,feature_template,f"sb_{config_name}.yaml")

    # if ":" in feature_template:
    #     template_name, version = feature_template.split(":")

    #     yaml_path = os.path.join(template_root,template_name.strip(),version.strip(),f"sb_{config_name}.yaml")
    # else:
    #     version = "lts"
    #     yaml_path = os.path.join(template_root,feature_template,version,f"sb_{config_name}.yaml")
    
    if not os.path.exists(yaml_path):
        raise ValueError(f"Template {yaml_path} does not exist")
    
    template_yaml = read_yaml(yaml_path)

    print("Installing Feature dependencies")
    install_packages_in_venv(venv,template_yaml['dependencies'])


import os
import asyncio

from speedbuild.utils.parsers.python.parser import PythonBlockParser
from speedbuild.utils.parsers.javascript_typescript.jsParser import JsTxParser
from speedbuild.frameworks.express.extraction.jsDep import getChunkDependencies
from speedbuild.frameworks.django.extraction.feature_dependencies import getBlockDependencies


def formatPythonOutput(chunks):
    read_data = []
    import_data = []

    for chunk in chunks:
        if chunk.startswith("from ") or chunk.startswith("import "):
            import_data.append(chunk)
        else:
            dependencies_data = []
            dependencies = getBlockDependencies(chunk,chunks)
            for dep in dependencies:
                if not dep['packagePath'].startswith("django."):
                    data = {
                        "name":dep['imports'],
                        "source":dep['packagePath']
                    }
                    if data not in dependencies_data:
                        dependencies_data.append(data)
                        
            read_data.append({
                "name":f"chunk_{len(read_data)+1}",
                "code":chunk,
                "dependencies":dependencies_data
            })

    if len(import_data) > 0:
        read_data.insert(0,{
            "name":"import_chunk",
            "code":"\n".join(import_data)
        })

    return read_data

def getDepSource(import_data,dep):
    for i in import_data:
        elememt = import_data[i]
        for entry in elememt:
            if entry['dep'] == dep:
                return i
    return None

def formatJavascripttOutput(imports,chunks,chunk_names,import_deps):
    read_data = []

    for chunk in chunks:
        dependencies_data = []
        dependencies = getChunkDependencies(chunk,chunk_names,False)

        for dep in dependencies:
            source = getDepSource(import_deps,dep)
            if source == None:
                source = "."

            data = {
                "name":dep,
                "source":source
            }
            dependencies_data.append(data)

        read_data.append({
            "name":f"chunk_{len(read_data)+1}",
            "code":chunk,
            "dependencies":dependencies_data
        })

    read_data.insert(0,{
        "name":"import_chunk",
        "code":"\n".join(imports)
    })

    return read_data


def formatOutput(chunks,file_type):
    if file_type == "python":
        return formatPythonOutput(chunks)

    return formatJavascripttOutput(chunks)    

def read_file(file_name : str):
    """
    Get File content.

    Args:
        file_name (str) : absolute path of file

    Returns:
        List of chunks of file content
    """
    if not os.path.exists(file_name):
        return None
    
    
    if file_name.endswith(".py"):
        with open(file_name,"r") as file:
            chunks = PythonBlockParser().parse_code(file.read())
            return formatPythonOutput(chunks)
    else:
        imports,chunks,variableToChunkMapping,import_deps = asyncio.run(JsTxParser().parse_code(file_name))
        return formatJavascripttOutput(imports,chunks,variableToChunkMapping,import_deps)

    # print(chunks)
    


# if __name__ == "__main__":
#     file_name = "/home/attah/.sb/environment/express/controllers/contact_controller.js"
#     data = read_file(file_name,"js")

#     print(data)

#     # for chunk in data:
#     #     print(chunk['name'])
#     #     print(chunk['code'])
#     #     if chunk['name'] != "import_chunk":
#     #         print("\n",chunk['dependencies'])
#     #     print("\n","-"*30,"\n")


import os
import zipfile
from pathlib import Path

from speedbuild.utils.cleanup.template.clear_folder import clear_folder
from speedbuild.utils.template.file import getCurrentDjangoFiles 
from speedbuild.utils.parsers.python.parser import PythonBlockParser 

from ..extraction.extract import getTemplateFileNames
from ..extraction.feature_dependencies import removeDuplicates
from ..extraction.write_dependency import sortFile, writeToFile

user_home = str(Path.home())


def getAppFileContent(appName,fileName,project_path):
    """
    Reads a Python file, separates import statements from the rest of the code, and returns them as two lists.

    Args:
        appName (str): The name of the application (used to construct the file path).
        fileName (str): The path to the file to be read.
        project_path (str): The root path of the project.

    Returns:
        list: A list containing two elements:
            - imports (list): A list of import statements found in the file.
            - code (list): A list of code blocks that are not import statements.

    Note:
        This function uses PythonBlockParser().parse_code to split the file content into code chunks.
    """
    imports = []
    code = []

    # fileToUpdate = fileName.split("/")[-1]
    # filePath = f"{project_path}/{appName}/{fileToUpdate}"

    with open(fileName, "r") as file:
        data = file.read()
        data = PythonBlockParser().parse_code(data)
        for chunk in data:
            if chunk.startswith("import ") or chunk.startswith("from "):
                # individualImports = getIndividualImports(chunk)
                imports.append(chunk)
            else:
                code.append(chunk)

    return [imports,code]


def processFile(fileName,appName,project_path,template_path,django_root=None,processed_file_path={}):
    imports = []
    code = []

    #get new feature code
    with open(fileName,"r") as file:
        data = file.read()
        data = PythonBlockParser().parse_code(data)

        for chunk in data:
            if chunk.startswith("import ") or chunk.startswith("from "):
                # individualImports = getIndividualImports(chunk)
                imports.append(chunk)
            else:
                code.append(chunk)
    
    #get existing file content
    fileImports, fileCode = getAppFileContent(appName,fileName,project_path)

    # old implementation
    fileCode.extend(code)
    fileImports.extend(imports)
    fileCode = removeDuplicates(fileCode)
    fileImports = removeDuplicates(fileImports)

    fileToUpdate = os.path.basename(fileName)
    filePath = os.path.join(project_path,appName)

    "get all files in filePath"

    # make this dynamic, get all python file in app folder
    django_files = getCurrentDjangoFiles(filePath)#["models.py","urls.py","views.py","admin.py"]

    if "sb_app" not in fileName: # also check if filename is custom django names else write file in sb_utils folder

        new_file_path = fileName.replace(template_path,"").rstrip("/")
        if new_file_path.startswith("/"):
            new_file_path = new_file_path[1:]

        write_to_project_root = len(new_file_path.split("/")) > 1

        if write_to_project_root:

            if fileToUpdate not in django_files:
                filePath = os.path.join(filePath,"sb_utils") 
                #  update import here
                # writing to sb utils
                new_imports = []
                for line in fileImports:
                    path,dependency = line.split("import ")
                    path = path.replace("from","").replace(".sb_utils","").strip()
                    if path.startswith("."):
                        path = path[1:]


                    if f"{path}.py" in django_files:
                        new_imports.append(f"from {appName}.{path} import {dependency}")
                        continue

                    if "sb_utils" in filePath:
                        line = line.replace(".sb_utils","")

                    new_imports.append(line)

                fileImports = new_imports
        else:
            if django_root:
                filePath = os.path.join(project_path,django_root)
            # settings files
            # TODO files to add to the main django folder

        if not os.path.exists(filePath):
            os.makedirs(filePath)

    
    # update imports to remove unnecessary imports
    new_imports = []
    for line in fileImports:
        if ".sb_utils" in line:
            path,dependency = line.split("import")
            path = path.replace("from","").strip()
            path = path.replace(".sb_utils.","").strip()
            # print("sb_utils path is ", path)

            if f"{path}.py" in django_files:
                if f"{path}.py" != fileToUpdate:
                    new_imports.append(f"from .{path} import {dependency}")
                continue

            elif path in processed_file_path.keys():
                main_path = processed_file_path[path]
                new_imports.append(f"from {main_path}.{path} import {dependency}")
                continue

        new_imports.append(line)
    fileImports = new_imports

    importAsString = "\n".join(fileImports)
    codeAsString = "\n".join(fileCode)
    fileContent = importAsString + "\n\n" + codeAsString

    writeToFile(filePath,fileContent,fileToUpdate)
    if fileToUpdate == "models.py":
        sortFile(os.path.join(filePath,fileToUpdate))


def getTemplateFilteredFiles(template_unpack_path,feature_name):
    root_files = []
    filtered_files = []
    files = getTemplateFileNames(template_unpack_path)
 
    for i in files:
        if i not in ["settings.py", f"sb_{feature_name}.yaml"] and i.endswith("md") == False:
            if len(i.split("/")) == 1:
                # add to root_files also add to filtered_files
                root_files.append(i)
                filtered_files.append(i)

            elif len(i.split("/")) > 1:
                filtered_files.append(i)

    filtered_files = sorted(filtered_files, key=lambda x: not x.startswith('.sb_utils')) 
    return [filtered_files,root_files]

        
def getFeatureFromTemplate(template_path,project_root,template_name,django_root,app_name):
    """
    1) Ask for which app to implement feature. -- done
    2) Ask for Customization prompt and customize code. -- done
    3) Copy code and save in the right files (do proper refrencing).
    """
    feature_name = [word for word in template_path.split("/") if ".zip" not in word][-1]
    path_to_template = os.path.join(user_home,".sb","sb_extracted")
    
    if not os.path.exists(path_to_template):
        os.makedirs(path_to_template, exist_ok=True)

    processed_file_path = {}

    # print("\n\n------- Generating Feature -------\n\n")

    # print(f"Getting template from {template_path}")

    template_unpack_path = os.path.join(path_to_template,feature_name)
    os.makedirs(template_unpack_path, exist_ok=True)

    # unpack
    # TODO : clear previous zip content, before unzipping again
    clear_folder(template_unpack_path)
    
    # print("Unpacking template")
    with zipfile.ZipFile(template_path, 'r') as zip_ref:
        zip_ref.extractall(template_unpack_path)

    # Get and read template configuration file
    feature_name = feature_name.replace("speed_build_","")
    template_root = os.path.join(path_to_template,template_name)

    filtered_files = []
    root_files = []

    files = getTemplateFileNames(template_unpack_path)

    for i in files:
        if i not in ["settings.py", f"sb_{feature_name}.yaml"] and i.endswith("md") == False:
            if len(i.split("/")) == 1:
                root_files.append(i)

            elif len(i.split("/")) > 1:
                filtered_files.append(i)

    filtered_files = sorted(filtered_files, key=lambda x: not x.startswith('.sb_utils')) 

    for root_file in root_files:
        path_name = root_file.replace(".py","").strip()
        processed_file_path[path_name] = django_root 

    root_files.extend(filtered_files) # merge files NB, made sure root files appear first
    filtered_files = root_files

    # TODO: print feature description, along with dependencies and fields

    # TODO : get filtered files again so u get new files llm created
    filtered_files,root_files = getTemplateFilteredFiles(template_unpack_path,feature_name)

    for file in filtered_files:
        file_path = os.path.join(template_root,file)
        # print(file_path)
        processFile(file_path,app_name,project_root,template_root,django_root,processed_file_path)
    
    return [app_name,template_root]



DEBUG = False

if DEBUG:
    url_address = "http://127.0.0.1:8000/"
    ws_address = "ws://127.0.0.1:8000/ws/"
    app_address = "http://localhost:3000/"
else:
    url_address = "https://backend.speedbuild.dev/"
    ws_address = "wss://backend.speedbuild.dev/ws/"
    app_address = "https://app.speedbuild.dev/"


import asyncio
from .read_file import read_file
from speedbuild.utils.parsers.javascript_typescript.jsParser import JsTxParser
from speedbuild.utils.parsers.python.parser import PythonBlockParser, indentCode


def unindentCode(code):
    """
    Split and partially unindent a block of code into separate "blocks" based on top-level lines.
    This function processes a multiline string `code` and returns a list of text blocks. Each returned
    block is a contiguous group of input lines (joined with newline characters) accumulated until a
    top-level (no-leading-space) line is encountered, at which point the current accumulated block is
    closed and a new block starts. Blank lines in the input are preserved as single newline entries
    within blocks.
    Behavior details:
    - Input is split by newline characters into individual lines.
    - Purely blank lines (lines for which line.strip() == "") are preserved as a single newline and
        appended to the current block (or used to separate blocks).
    - For lines processed before any top-level (indent_count == 0) line has been seen:
            - Lines with an indentation less than 4 spaces have their surrounding whitespace stripped
                (equivalent to line.strip()).
            - Lines with an indentation of 4 or more spaces have the first four spaces removed (only one
                indentation level is removed).
        This "pre-top-level" unindent behavior stops once a top-level line is seen.
    - When a top-level line (no leading spaces) is encountered:
            - If any lines have been accumulated since the last block boundary, that accumulated chunk is
                finalized and appended to the result list.
            - A flag is set so subsequent lines are not unindented by the special rules above.
    - After processing all lines, any remaining accumulated lines form the final block appended to the
        result.
    Parameters:
    - code (str): Multiline string containing the code to split and partially unindent.
    Returns:
    - List[str]: A list of string blocks. Each element is a block constructed by joining the cleaned
        lines for that block with newline characters.
    Notes and caveats:
    - The function treats indentation as spaces only. Tab characters are not specially handled and may
        produce unexpected results.
    - The function removes at most four leading spaces from lines in the initial (pre-top-level)
        region; it does not normalize arbitrary indentation levels beyond that.
    - Blank lines are represented as "\n" within blocks; consecutive blank lines will be preserved as
        separate entries in the accumulated text for that block.
    - The implementation assumes non-empty lines that contain only spaces are handled by the early
        blank-line check (so they are not iterated character-by-character).
    Example:
    >>> code = "    def foo():\\n        pass\\n\\nclass Bar:\\n    pass"
    >>> unindentCode(code)
    ["def foo():\\n    pass\\n\\n", "class Bar:\\n    pass"]
    """

    blocks = []
    cleaned = []
    has_sub_chunk = False
    lines = code.split("\n")

    for line in lines:

        if len(line.strip()) == 0:
            cleaned.append("\n")
            continue

        indent_count = 0
        while line[indent_count] == " ":
            indent_count += 1

        if indent_count == 0:
            if len(cleaned) > 0:
                blocks.append("\n".join(cleaned))
                cleaned = []
                has_sub_chunk = True

        elif indent_count < 4:
            if not has_sub_chunk:
                line = line.strip()
        else:
            if not has_sub_chunk:
                _,line = line.split("    ",1)
        
        cleaned.append(line)

    if len(cleaned) > 0:
        blocks.append("\n".join(cleaned))

    return blocks


# Source of my problem
def processJsChunk(code):
    lines = code.split("\n")

    if len(lines) < 2:
        return code
    
    first_line = lines[0]
    last_line = lines[-1]

    code = lines[1:len(lines)-1]

    # whitespace = 0
    # while code[0][whitespace] == " ":
    #     whitespace += 1

    # if whitespace > 0:
    #     for i in range(0,len(code)):
    #         line = code[i]
    #         # print("line info",line.split(" "*whitespace,1))
    #         chunks = line.split(" "*whitespace,1)
    #         if len(chunks)  > 1:
    #             code[i] = chunks[1]

    code = "\n".join(code)

    _,chunks,_,_ = asyncio.run(JsTxParser().parse_code(code,False,False,True))

    chunks.insert(0,first_line)
    chunks.append(last_line)

    return chunks

def processChunk(code):
    """
    Breaks down Python code into logical chunks while preserving data structure definitions.
    This function splits code into chunks based on logical breaks and handles special cases
    for data structures (lists, dictionaries, tuples). It maintains proper indentation and
    structure of the original code.
    Args:
        code (str): A string containing Python code to be broken into chunks.
    Returns:
        list or str: If the code can be broken down, returns a list of code chunks.
                    If the code is a single line, returns the original code string.
    Examples:
        >>> code = "def example():\\n    x = [1,\\n    2,\\n    3]"
        >>> breakChunk(code)
        ['def example():', '    x = [', '    1,', '    2,', '    3', ']']
        >>> code = "single_line = 42"
        >>> breakChunk(code)
        'single_line = 42'
    """

    if len(code.split("\n")) < 2:
        return code
    
    first_line, code = code.split("\n",1)

    chunk_symbol = None
    is_data_structure = False
    data_structure_closing = ["]","}",")"]
    data_structure_starting = ["[","{","("]

    possible_opening = set(first_line).intersection(set(data_structure_starting))
    chunk_start = first_line.split(" ",1)[0]

    index = None
    
    if chunk_start not in ["def","class"] and len(list(possible_opening)) > 0:
        for char in possible_opening:
            if index == None:
                index = first_line.index(char)
                continue
            pos = first_line.index(char)
            if pos < index:
                index = pos
        
        chunk_symbol = first_line[index]
        code = f"{first_line[index+1:]}{code}"

        first_line = first_line[:index + 1]
        is_data_structure = True

    chunks = unindentCode(code)

    if len(chunks) > 1 and not is_data_structure:
        chunks[0] = first_line + "\n" + indentCode(chunks[0])
        return chunks
    
    splits = PythonBlockParser().parse_code(chunks[0])
    chunks = []
    for chunk in splits:
        chunks.append(indentCode(chunk))

    chunks.insert(0,first_line)

    if is_data_structure:
        close_symbol = data_structure_closing[data_structure_starting.index(chunk_symbol)]
        
        if chunks[-1].endswith(","):
            chunks[-1] = chunks[-1][:len(chunks[-1])-1]

        if chunks[-1].endswith(close_symbol):
            chunks[-1] = chunks[-1][:len(chunks[-1])-1]

        chunks.append(close_symbol)
    
    return chunks

def removeLeadingSpaces(code):
    count = 0
    cleaned = []
    while code[count] == " ":
        count += 1

    if count > 0:
        lines = code.split("\n")
        for line in lines:
            cleaned.append(line[count:])
        
        code = "\n".join(cleaned)

    return code,count

def addLeadingWhiteSpaces(code,count):
    if count == 0:
        return code

    return {"name":code['name'],"code":" "*count + code['code']}

def breakChunk(code,name,code_type="python"):
    # TODO : make dynamic, add leading spaces in some cases
    endWithComa = False

    if code.endswith(","):
        endWithComa = True
        code = code[:len(code)-1]

    formatted = []
    code, whiteSpaceCount = removeLeadingSpaces(code)

    if code_type == "python":
        result = processChunk(code)

        if endWithComa and len(result) > 0:
            result[-1] = result[-1] + ","

    else:
        result = processJsChunk(code)

    if isinstance(result,str):
        formatted.append({
            "name":f"{name}",
            "code":result
        })

    else:
        for i in result:
            formatted.append({
                "name":f"{name}_sub_{len(formatted)+1}",
                "code":i 
            })

    return formatted,whiteSpaceCount


def getLayerBreakDown(name):
    if name == None:
        return []
    
    names = name.split("_sub_")
    layers = [names[0]]

    for i in range(1,len(names)-1):
        prev = layers[i-1]
        layers.append(f"{prev}_sub_{names[i]}")
    
    return layers
    
def getChunk(name,chunks):
    for chunk in chunks:
        if chunk['name'] == name:
            return chunk
    return None
    
def getLayerCode(layers,layer_chunks,file_type):

    print(layers)

    while len(layers) > 0:
        layer = layers.pop(0)
        chunk = getChunk(layer,layer_chunks)
        layer_chunks,_ = breakChunk(chunk['code'],layer,file_type)

    return layer_chunks


def break_chunk(file_name: str, chunk_name : str):
    """
    Break code chunks into sub chunks

    Args:
        file_name (str) : absolute path of the file holding the chunk
        chunk_name (str) : name of chunk we want to break into sub chunks

    Returns:
        Successful : a list os sub chunks
        Failure : Error message. 
    """

    file_type = "js"
    if file_name.endswith(".py"):
        file_type = "python"
    
    file_content = read_file(file_name)
    layers = getLayerBreakDown(chunk_name)

    if file_content is None:
        return f"Could not read file {file_name}, because it does not exist"
    
    chunks = getLayerCode(layers,file_content,file_type)

    if chunks is not None:
        return chunks
    
    return f"Could find chunk with name {chunk_name} in file"


if __name__ == "__main__":
    chunks = break_chunk('/home/attah/.sb/environment/express/controllers/contact_controller.js',"chunk_1","js")
    print(chunks)


# if __name__ == "__main__":
#     code = """
#         {
#             "step": "Basic Info",
#             "completed": True,
#         },
#         {
#             "step": "Health Goals",
#             "completed": False,
#         },
#         {
#             "step": "LifeStyle",
#             "completed": False,
#         },
#         {
#             "step": "Medical History",
#             "completed": False,
#         },
#         {
#             "step": "Digestive History",
#             "completed": False,
#         },
#         {
#             "step": "Diet History",
#             "completed": False,
#         },
#         {
#             "step": "Reproductive Health",
#             "completed": False,
#         },
#         {
#             "step": "Emergency Contact",
#             "completed": False,
#         },
#     """.strip()

#     result = breakChunk(code,"hello")
#     if isinstance(result,str):
#         print(result)
#     else:
#         for i in result:
#             print(i,"\n","-"*30)


import os

def getTemplateFileNames(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root or  "migrations" in root or  "__pycache__" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            if not absolute_path.endswith(".yaml"): #TODO : remove this in the future
                file_paths.append(absolute_path)
    return file_paths


import os
import json
from pathlib import Path

user_home = str(Path.home())
config_path = os.path.join(user_home,".sb","config.json")

supported_models = {
    "openai":["gpt-4o"],
    "anthropic":[],
}

def get_sb_config():
    if not os.path.exists(config_path):
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        with open(config_path,"w") as file:
            file.write("{}")

        return {}

    with open(config_path,"r") as file:
        data = json.loads(file.read())
        return data

def save_config(config):
    if not os.path.exists(config_path):
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        
    with open(config_path, "w") as file:
        json.dump(config, file, indent=4)

def get_config():
    config = get_sb_config()
    # print(json.dumps(config, indent=4))
    return config

def setAPIKey(provider,api_key):
    if provider not in supported_models.keys():
        raise ValueError(f"Please choose from our LLM supported providers {list(supported_models.keys())}")
    
    config = get_sb_config()

    keys_dictionary = {}
    if "llm_keys" in config.keys():
        keys_dictionary = config['llm_keys']

    if len(api_key) == 0 and provider in keys_dictionary.keys():
        del keys_dictionary[provider]
    else:
        keys_dictionary[provider] = api_key

    
    config['llm_keys'] = keys_dictionary

    save_config(config)

def updateDefaultLLMModel(provider,model_name):
    config = get_sb_config()

    if provider not in supported_models.keys():
        raise ValueError("Please choose from our supported providers ",list(supported_models.keys()))
    
    if model_name not in supported_models[provider]:
        raise ValueError(f"Supported Models for provider : {provider} are {supported_models[provider]}")
    
    config['default_model'] = model_name
    config['model_provider'] = provider
    save_config(config)

def updateAuthToken(token_value):
    config = get_sb_config()

    keys_dictionary = {}
    if "jwt_tokens" in config.keys():
        keys_dictionary = config['jwt_tokens']

    if len(token_value) == 0:
        del config['jwt_tokens']
    else:
        config['jwt_tokens'] = token_value

    save_config(config)


import ast
from typing import List, Tuple


def extract_views_from_urls(urls_file_path: str) -> List[Tuple[str, str]]:
    print(f"Extracting views from {urls_file_path}")
    with open(urls_file_path, "r") as f:
        data = f.read()
        data = data.replace(".as_view()", "")

        tree = ast.parse(data, filename=urls_file_path)

        view_names = []
        import_alias_map = {}

        # Step 1: Capture import statements to trace views
        for node in ast.walk(tree):
            if isinstance(node, ast.ImportFrom):
                module = node.module
                if module == None:
                    module = ""
                for alias in node.names:
                    import_alias_map[alias.asname or alias.name] = f"{module}.{alias.name}"
            elif isinstance(node, ast.Import):
                for alias in node.names:
                    import_alias_map[alias.asname or alias.name] = alias.name

        # Step 2: Find calls to `path()` or `re_path()`
        for node in ast.walk(tree):
            
            if isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id in ("path", "re_path"):
                
                if len(node.args) >= 2:
                    view = node.args[1]
                    if isinstance(view, ast.Attribute):

                        # Something like views.HomeView
                        value_id = view.value.id if isinstance(view.value, ast.Name) else None

                        attr = view.attr
                        full_path = import_alias_map.get(value_id)
                        if full_path:
                            view_names.append(f"{full_path}.{attr}")
                    elif isinstance(view, ast.Name):
                        # Direct function/class reference
                        view_path = import_alias_map.get(view.id)
                        if view_path:
                            view_names.append(view_path)

    return view_names


import os
import json
from typing import List
from pathlib import Path

from .prompt import get_system_prompt
from .get_files import getTemplateFileNames

from .tools.read_file import read_file
from .tools.update_file import SingleUpdateAction, updateFile
from .tools.break_chunk import break_chunk
from .tools.execute_command import executeCommand

from speedbuild.agents.config import getLLMConfig
from speedbuild.utils.cli.output import StatusManager

from langchain.chat_models import init_chat_model
from langgraph.checkpoint.memory import InMemorySaver
from langgraph.prebuilt import ToolNode, tools_condition
from langgraph.graph import START,StateGraph, MessagesState
from langchain_core.messages import SystemMessage, HumanMessage


class SpeedBuildAgentV2():
    def __init__(self,framework,debugging=False):
        self.first_message = True
        self.framework = framework
        self.debugging = debugging

        model_provider, model = getLLMConfig()
        
        model = init_chat_model(model_provider=model_provider, model=model)
        self.model = model.bind_tools([self.run_read_file,self.run_break_chunk,self.run_update_file])

        self.logger = StatusManager()

        if framework == "django":
            self.test_environment = os.path.join(str(Path.home()),".sb","environment","django","speedbuild_project")
        else:
            self.test_environment = os.path.join(str(Path.home()),".sb","environment","express")

        template_files = None if debugging else getTemplateFileNames(self.test_environment)
        self.system_prompt = get_system_prompt(framework,template_files,debugging) #set system prompt

        self.build_graph()

    def run_read_file(self,file_name:str,description:str):
        """
        Get File content.

        Args:
            file_name (str) : absolute path of file
            description (str) : short single sentence description of action

        Returns:
            List of chunks of file content
        """
        self.logger.update_status(description)
        return read_file(file_name)

    def run_break_chunk(self,file_name:str,chunk_name:str,description:str):
        """
        Break code chunks into sub chunks

        Args:
            file_name (str) : absolute path of the file holding the chunk
            chunk_name (str) : name of chunk we want to break into sub chunks
            description (str) : short single sentence description of action

        Returns:
            Successful : a list os sub chunks
            Failure : Error message. 
        """
        self.logger.update_status(description)
        return break_chunk(file_name,chunk_name)
    
    def run_update_file(self,filename:str,description:str, actions : List[SingleUpdateAction] = []):
        """
        Update File; write, remove or replace code in file

        Args:
            filename (str) : absolute path to file we are updating
            description (str) : short single sentence description of action
            actions (List[SingleUpdateAction]) : a list of update action to perform on file

        Returns:
            Confirmation if update was successful or not
        """
        self.logger.update_status(description)
        return updateFile(filename,actions)


    def chatAI(self,state):
        res = self.model.invoke(state['messages'])
        # print(res)
        return {"messages":res}
    
    def build_graph(self):
        builder = StateGraph(MessagesState)
        builder.add_node("chat",self.chatAI)
        builder.add_node("tools",ToolNode([
            self.run_read_file,
            self.run_break_chunk,
            self.run_update_file
        ]))

        builder.add_edge(START,"chat")
        builder.add_conditional_edges("chat",tools_condition)
        builder.add_edge("tools","chat")
        self.graph = builder.compile(checkpointer=InMemorySaver())

    def call_agent(self,msg,history):
        messages = [HumanMessage(content=msg)]
        if self.first_message:
            self.first_message = False
            messages = [
                SystemMessage(content=self.system_prompt),
                HumanMessage(content=msg)
            ]
        return self.graph.invoke({"messages":messages},{"configurable": {"thread_id":history}})
    
    def convertJson(self,data):
        data = data.replace("`json","").replace("`","").strip()
        return json.loads(data)
    
    def run(self,query,history_id="1"):
        self.logger.start_status("Debugging" if self.debugging else "Customizing")
        while True:
            response = self.call_agent(query,history_id)
            last = response['messages'][-1].content
            formatted = self.convertJson(last)
        
            if 'command' in formatted.keys():
                self.logger.update_status(formatted['message'])
                query = executeCommand(formatted['command'],self.framework)
            elif "type" in formatted.keys() and formatted['type'] == "error":
                # Agent failed
                self.logger.stop_status("Agent Unsuccessful")
                raise ValueError(formatted['message'])
            else:
                self.logger.stop_status("Speedbuild Agent finished Successfully")
                return formatted['message']


import os
import json
import shutil
import requests
import tempfile
import subprocess
import importlib.util
import concurrent.futures
from importlib import metadata

from rich.live import Live
from rich.table import Table
from rich.console import Console
from rich.layout import Layout
from rich.spinner import Spinner

from speedbuild.utils.cli.output import StatusManager 
from speedbuild.auth.serverCall import manageServerCall


MAX_WORKERS = 5  # Reduced from 5 to avoid resource exhaustion
TIMEOUT = 120  # Seconds to wait before considering a package installation hung

def load_django_settings(settings_path):
    """Loads the Django settings module dynamically."""
    spec = importlib.util.spec_from_file_location("settings", settings_path)
    settings_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(settings_module)
    return settings_module

def create_temp_env():
    """Creates a temporary virtual environment and returns its path."""
    temp_dir = tempfile.mkdtemp()
    venv_path = os.path.join(temp_dir, "venv")
    try:
        # Add timeout to avoid hanging
        result = subprocess.run(
            ["python", "-m", "venv", venv_path], 
            check=True, 
            timeout=30  # 30 seconds should be enough for venv creation
        )
        return temp_dir, venv_path
    except (subprocess.TimeoutExpired, subprocess.SubprocessError) as e:
        # print(f"Error creating virtual environment: {str(e)}")
        shutil.rmtree(temp_dir)
        raise

def get_site_packages_path(venv_path):
    """Returns the site-packages path for the virtual environment."""
    python_bin = os.path.join(venv_path, "Scripts" if os.name == "nt" else "bin", "python")
    try:
        output = subprocess.run(
            [python_bin, "-c", "import site; print(site.getsitepackages()[0])"],
            capture_output=True,
            text=True,
            timeout=10  # 10 seconds timeout
        )
        return output.stdout.strip()
    except subprocess.TimeoutExpired:
        print(f"Timeout getting site-packages path for environment: {venv_path}")
        raise

def install_package(venv_path, package_name):
    """Installs a package in the temporary virtual environment."""
    pip_bin = os.path.join(venv_path, "Scripts" if os.name == "nt" else "bin", "pip")
    try:
        # print(f"Starting installation of {package_name}...")
        process = subprocess.run(
            [pip_bin, "install", package_name], 
            check=True, 
            timeout=TIMEOUT,  # Timeout to prevent hanging
            capture_output=True,
            text=True
        )
        # print(f"Detected {package_name}")
        return True
    except subprocess.TimeoutExpired:
        # print(f"Installation of {package_name} timed out after {TIMEOUT} seconds")
        return False
    except subprocess.SubprocessError as e:
        # print(f"Error installing {package_name}: {str(e)}")
        return False

def detect_installed_apps(site_packages_path):
    """Detect likely Django apps in site-packages."""
    try:
        apps = []
        for item in os.listdir(site_packages_path):
            full_path = os.path.join(site_packages_path, item)

            # Ignore metadata folders
            if item.endswith(".dist-info") or item.endswith(".egg-info"):
                continue

            # Include real directories
            if os.path.isdir(full_path):
                # Check for __init__.py to confirm it's a module
                # if os.path.isfile(os.path.join(full_path, "__init__.py")):
                #     apps.append(item)
                apps.append(item)
        return apps
    except Exception as e:
        print(f"Error detecting installed apps: {str(e)}")
        return []

def process_single_package(package_name):
    """Process a single package in its own virtual environment."""
    # print(f"Processing package: {package_name}")
    
    temp_dir = None
    try:
        temp_dir, venv_path = create_temp_env()
        site_packages = get_site_packages_path(venv_path)
        
        success = install_package(venv_path, package_name)
        if not success:
            # print(f"Installation failed for {package_name}")
            return package_name, []
            
        installed_apps = detect_installed_apps(site_packages)
            
        return package_name, installed_apps
    except Exception as e:
        print(f"Error processing {package_name}: {str(e)}")
        return package_name, []
    finally:
        # Cleanup temp environment
        if temp_dir and os.path.exists(temp_dir):
            try:
                shutil.rmtree(temp_dir)
                # print(f"Cleaned up environment for {package_name}")
            except Exception as e:
                pass
                # print(f"Failed to clean up {temp_dir}: {str(e)}")


def get_package_version(package_name):
    try:
        return metadata.version(package_name)
    except metadata.PackageNotFoundError:
        return "unknown"

def get_django_app_from_packages_parallel(package_names, max_batch_size=10):
    """
    TODO : here we dance

    Process multiple package names in parallel to retrieve Django applications.
    This function processes a list of package names to identify Django applications within them,
    utilizing parallel processing for efficiency. It implements caching to avoid reprocessing
    previously analyzed packages.
    Args:
        package_names (list): A list of package names to process.
        max_batch_size (int, optional): Maximum number of packages to process in parallel. 
            Defaults to 10.
    Returns:
        dict: A dictionary mapping package names to lists of discovered Django applications.
            For packages that fail processing or timeout, an empty list is returned.
    Example:
        >>> packages = ['django-allauth', 'django-rest-framework']
        >>> results = get_django_app_from_packages_parallel(packages)
        >>> print(results)
        {'django-allauth': ['allauth', 'allauth.account'], 'django-rest-framework': ['rest_framework']}
    Note:
        - Uses threading for parallel processing
        - Implements timeout handling to prevent hanging
        - Caches results for future lookups
        - Processes packages in batches to manage resource usage
    """

    
    results = {}
    logger = StatusManager()
    packages_to_process = list(package_names)  
    
    logger.start_status("Getting Installed packages")

    # Process in smaller batches to prevent resource exhaustion
    for i in range(0, len(packages_to_process), max_batch_size):
        batch = packages_to_process[i:i+max_batch_size]
        # print(f"Processing batch of {len(batch)} packages")

        batch_with_version = [
            f"{i}=={get_package_version(i)}"
            for i in batch
        ]

        # make server request to get saved packages
        try:
            success,res = manageServerCall("POST","api/packages/",{
                "action":"get",
                "packages":json.dumps(batch_with_version)
            })
        except requests.exceptions.ConnectionError:
            success = False

        if success:
            recieved_packages = res['packages']
            recieved_packages_names = [name.split("==")[0].strip() for name in recieved_packages.keys()]
            batch = [name for name in batch if name not in recieved_packages_names ]

            for pkg in recieved_packages:
                results[pkg.split("==")[0]] = recieved_packages[pkg] #remove version from recieved_packages before adding to results
        
        new_packages = {}

        with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
            future_to_package = {
                executor.submit(process_single_package, pkg): pkg 
                for pkg in batch
            }
            
            for future in concurrent.futures.as_completed(future_to_package):
                try:
                    package_name, installed_apps = future.result(timeout=TIMEOUT + 30)
                    package_version = get_package_version(package_name)
                    new_packages[f"{package_name}=={package_version}"] = {"pkg":installed_apps,"version":package_version}#installed_apps

                    logger.update_status("Installed Package Detected : "+package_name)
                    # add to reults
                    results[package_name] = {"pkg":installed_apps,"version":package_version}
                    
                except concurrent.futures.TimeoutError:
                    package = future_to_package[future]
                    print(f"Processing timed out for package: {package}")
                    # results[package] = {"pkg":[],"version":None}
                except Exception as e:
                    package = future_to_package[future]
                    # print(f"Exception for package {package}: {str(e)}")
                    # results[package] = {"pkg":[],"version":None}

        # push by batch to the server
        if len(new_packages.keys()) > 0:
            try:
                success,res = manageServerCall("POST","api/packages/",{
                    "action":"save",
                    "packages":json.dumps(new_packages)
                })
            except requests.exceptions.ConnectionError:
                pass
    logger.stop_status("Installed Packages Collected")
    return results

def get_installed_packages(venv_path):
    """Get list of installed packages in the given virtual environment."""
    pip_path = os.path.join(venv_path, "bin", "pip")  # Linux/macOS

    if os.name == "nt":
        pip_path = os.path.join(venv_path, "Scripts", "pip.exe")  # Windows

    try:
        result = subprocess.run(
            [pip_path, "list"], 
            capture_output=True, 
            text=True,
            timeout=30  # 30 seconds timeout
        )
        lines = result.stdout.splitlines()[2:]  # Skip headers
        packages = {line.split()[0] for line in lines}  # Extract package names
        return packages
    except subprocess.TimeoutExpired:
        print("Timeout getting installed packages")
        return set()
    except Exception as e:
        print(f"Error getting installed packages: {str(e)}")
        return set()

def getDjangoAppsPackage(settings_path, venv):
    """
    Maps Django installed apps to their corresponding Python packages.
    This function analyzes the Django settings file to find installed apps and determines
    which Python packages provide those apps. It handles both direct app imports and
    regular Python package imports.
    Args:
        settings_path (str): Path to the Django settings.py file
        venv (str): Path to the Python virtual environment to analyze
    Returns:
        dict: A mapping of Django app names to lists of package names that provide them.
              Returns empty dict if settings file is not found or cannot be loaded.
              Format: {'app_name': ['package1', 'package2']}
    Raises:
        No exceptions are raised - errors are handled internally and logged to stdout
    Example:
        >>> getDjangoAppsPackage('/path/to/settings.py', '/path/to/venv')
        {'django.contrib.admin': ['django'],
         'django.contrib.auth': ['django'],
         'myapp': ['my-package']}
    """
    if not os.path.exists(settings_path):
        print(f"Error: Settings file '{settings_path}' not found!")
        return {}

    # Load Django settings
    try:
        settings_module = load_django_settings(settings_path)
        installed_apps = getattr(settings_module, "INSTALLED_APPS", [])
    except Exception as e:
        print(f"Error loading Django settings: {str(e)}")
        return {}

    # Get all installed packages
    packages = get_installed_packages(venv)
    
    # Process packages in parallel
    package_app_mapping = get_django_app_from_packages_parallel(packages)
    
    # Map Django apps to their packages
    app_mapping = {}
    for package, django_apps in package_app_mapping.items():
        for app in django_apps['pkg']:
            if app in installed_apps:  # or is a regular import
                if app in app_mapping:
                    app_mapping[app]['pkg'].append(package)
                else:
                    app_mapping[app] = {"pkg":[package]}

                    # check if version is not None or Unkmown
                    pkg_version = django_apps['version'] 
                    if pkg_version != None or pkg_version != "unknown":
                        app_mapping[app]['version'] = pkg_version 

    return [app_mapping,package_app_mapping]



"""
Add
Remove
Replace
"""

"""
Tasks
-----

1) group and process actions by layer                        -- done
2) breakCode output format should be the same as read_file   -- done
3) rework breakcode to handle list of expressions
4) validate if file operation was successful
5) validate if command execution was succesful
"""

"Process Layer by Layer"

from read_file import read_file
from break_chunk import breakChunk, getLayerBreakDown, getChunk, getLayerCode

from typing import List, Optional, TypedDict
from pydantic import Field

"""
    actions = [
        {
            "action":"add",
            "chunk":None,
            "code":<New code>
        },
        {
            "action":"remove",
            "chunk":<layer_chunk_name>,
            "code":None
        },
        {
            "action":"replace",
            "chunk":<layer_chunk_name>,
            "code":<New code>
        }
    ]
"""

class SingleUpdateAction(TypedDict):
    action : str = Field(description="action to carry out (add | remove | replace )")
    chunk : Optional[str] = Field(description = "name of chunk to use as anchor for update operation")
    code : Optional[str] = Field(description="New code to write to file")


def addLeadingWhiteSpaces(code,count):
    cleaned = []

    if count == 0:
        return code
    
    lines = code['code'].split("\n")
    for line in lines:
        cleaned.append(" "*count+line)

    cleaned = "\n".join(cleaned)

    return {"name":code['name'],"code":cleaned}

def processLayer(layers,layer_chunks,items,file_type="python",root_operation=False):

    if root_operation or len(layers) == 0: 

        # we are at the final layer
        for item in items:  # TODO : process layer actions once
            action = item['action']

            code = item['code']

            if action == "add":
                if item['chunk'] != None:
                    add_after = getChunk(item['chunk'],layer_chunks)
                    
                    if add_after == None:
                        raise ValueError(f"Chunk with name {item['chunk']} not found in file")
                    
                    position = layer_chunks.index(add_after)

                    layer_chunks.insert(position,{"code":code,"name":"new"}) #insert at specified position
                else:
                    layer_chunks.append({"code":code,"name":"new"}) #add code to ending of file

            elif action == "remove" or action == "replace":

                chunk_to_remove = getChunk(item['chunk'],layer_chunks)

                if chunk_to_remove == None:
                    raise ValueError(f"Chunk with name {item['chunk']} not found in {layer_chunks}")
                position = layer_chunks.index(chunk_to_remove)
                if action == "remove":
                    layer_chunks.pop(position)
                else:
                    layer_chunks[position] = {"code":code,"name":layer_chunks[position]['name']}

            else:
                raise ValueError(f"action {action} not recognised")
            
        return layer_chunks 
    else:

        layers = list(layers)
        current = layers.pop(0)

        chunk = getChunk(current,layer_chunks)

        if chunk == None:
            raise ValueError("Could not find code in file") #could not find layer in chunks
        
        inner_layer,whitespaces = breakChunk(chunk['code'],current,file_type) #TODO Start here

        result = processLayer(layers,inner_layer,items,file_type)

        prev = result
        if isinstance(result,list):
            prev = result[0]
        else:
            result = [result]

        prev = prev['name'].split("_sub_")
        prev = "_sub_".join(prev[:len(prev)-1]) # get previous layer name

        result_with_identation = [addLeadingWhiteSpaces(i,whitespaces)['code'] for i in result]
        
        chunk_index = layer_chunks.index(getChunk(current,layer_chunks))

        result = {"code":"\n".join(result_with_identation),"name":current}

        layer_chunks[chunk_index] = result

        # remove and add
        return layer_chunks
   
def sortAndGroupByLayer(data):
    # Group by identical chunk_layer array
    groups = {}
    for d in data:
        key = tuple(d.get("chunk_layer", []))  # tuples are hashable, lists aren't
        groups.setdefault(key, []).append(d)


    # Sort each group by number of chunk_layer (descending)
    for k in groups:
        # groups[k] = sorted(groups[k], key=lambda d: len(d.get("chunk_layer", [])), reverse=True)
        actions = sorted(groups[k], key=lambda d: len(d.get("chunk_layer", [])), reverse=True)
        groups[k] = {"actions":actions,"root_chunk":actions[0]['root_chunk']}


    # Sort the groups themselves from largest to smallest chunk_layer count
    sorted_groups = dict(sorted(groups.items(), key=lambda item: len(item[0]), reverse=True))

    # print(sorted_groups, "here with\n")

    return sorted_groups


def addChunkLayersBreakdown(layer_data):
    """
    actions = [
        {
            "action":"add",
            "chunk_layer":[],
            "chunk":None,
            "code":<New code>
        },
        {
            "action":"remove",
            "chunk_layer":[],
            "chunk":<layer_chunk_name>,
            "code":None
        },
        {
            "action":"replace",
            "chunk_layer":[],
            "chunk":<layer_chunk_name>,
            "code":<New code>
        }
    ]
    """

    for i in range(0,len(layer_data)):
        current = layer_data[i]
        layers = getLayerBreakDown(current['chunk'])
        current['chunk_layer'] = layers
        current["root_chunk"] = False

        if len(layers) == 1 and current['chunk'] == layers[0]:
            current["root_chunk"] = True

    return layer_data


def updateFile(filename : str,actions : List[SingleUpdateAction] = []) -> str:

    """
    Update File; write, remove or replace code in file

    Args:
        filename (str) : absolute path to file we are updating
        actions (List[SingleUpdateAction]) : a list of update action to perform on file

    Returns:
        Confirmation if update was successful or not
    """

    file_type = "js"

    if filename.endswith(".py"):
        file_type = "python"
   
    actions = addChunkLayersBreakdown(actions)

    chunks = read_file(filename)

    if chunks == None:
        "TODO : create new file in the specified path"
        file_code = ""
        for step in actions:
            print(step['action'])
            if step['action'] != "add":
                return "Error : This is a new file : you cannot replace or remove code because its new"
            
            if len(file_code) == 0:
                file_code += step['code']
            else:
                file_code += f"\n\n{step['code']}"

    else:
        sorted_data = sortAndGroupByLayer(actions) #sort by layer, max to min

        layer_chunks = chunks

        for layer in sorted_data:

            layer_actions = sorted_data[layer]['actions']
            is_root = sorted_data[layer]['root_chunk']


            # if is_root:
            #     print("Found root chunk ", layer)
                        
            # check for root operations
            try:
                layer_chunks = processLayer(layer,layer_chunks,layer_actions,file_type,is_root)
                # if root_chunk: # is root operation
                #     layer_chunks = processLayer(layer,layer_chunks,layer_actions,True)
                # else:
                #     # current = getChunk(layer[0],chunks)
                #     # update = getChunk(layer[0],layer_chunks)
                #     # # TODO : verify update here
                #     # print("current ",current,"\n\n")
                #     # print("update ",update,"\n\n")
                #     layer_chunks = processLayer(layer,layer_chunks,layer_actions)  
            except ValueError as err:
                return f"Error : {err}"
           

        file_code = "\n\n".join([i['code'] for i in layer_chunks])

    # write update
    with open(filename,"w") as file:
        file.write(file_code)

    return "Update Successful" # If Successful



if __name__ == "__main__":
    # print(getLayerBreakDown("chunk_1"))
    filename = "/home/attah/.sb/environment/express/controllers/contact_controller.js"
    chunks = read_file(filename)
    # layers = getLayerCode(getLayerBreakDown("chunk_1_sub_5_sub_3"),chunks,"js")

    # for item in layers:
    #     print("name",item['name'])
    #     print(item['code'])
    #     print("\n","-"*30)

    actions = [
        {
            "action":"remove",
            "chunk":"chunk_1_sub_2",
            "code":"console.log('Lolo')"
        },
        # {
        #     "action":"add",
        #     "chunk":"chunk_2", # add before chunk
        #     "code":"# 'step':'I am the Batman',"
        # },
        # {
        #     "action":"add",
        #     "chunk":"chunk_3_sub_2_sub_2_sub_2", # replace chunk
        #     "code":"'step':'Jesus Baby',"
        # },
        # {
        #     "action":"add",
        #     "chunk":"chunk_3_sub_2_sub_2_sub_3", # replace chunk
        #     "code":"'step':'Jesus Baby',"
        # },
        # {
        #     "action":"add",
        #     "chunk":None, # replace chunk
        #     "code":"New code"
        # },
    ]

    res = updateFile(filename,actions)
    print(res)




import os
from pathlib import Path
from typing import List,Dict, Optional

from get_files import getTemplateFileNames
from speedbuild.frameworks.django.utils.var_utils import get_assigned_variables
from tools.read_file import read_file

from pydantic import BaseModel

class Chunk(BaseModel):
    name: str
    code: str
    dependencies: Optional[List[str]]

def load_all_chunks(files: List[str]) -> Dict[str, Chunk]:
    all_chunks = []

    for path in files:
        raw = read_file(path)

        for item in raw:

            # cid = item['name']

            # Create a globally unique ID for the chunk
            # global_id = f"{path}:::{cid}"

            all_chunks.append({"source":path, **item})

    return all_chunks

def sortChunks(chunks):
    processed = set()
    arranged = []
    prev_batch = []
    cycles = []

    while len(chunks) > 0:
        process_next_batch = []
        for chunk in chunks:
            # print("processing ",chunk['name'],"\n")
            if "dependencies" not in chunk.keys():
                arranged.append(chunk)
                continue
            
            add_to_process = True
            for dep in chunk['dependencies']:
                dep_source = dep['source']

                if dep_source in [".",".sb_utils",".sb_app"] and dep['name'] not in processed:
                    add_to_process = False
                    print(dep,"stuck")
                    chunk_name = get_assigned_variables(chunk['code'],True)
                    print(chunk_name,"\n\n")
                    break

            if add_to_process:
                arranged.append(chunk)
                chunk_name = get_assigned_variables(chunk['code'],True)
                print(chunk_name)
                processed.add(chunk_name)
            else:
                process_next_batch.append(chunk)

        # if prev_batch == process_next_batch:
        #     # cycle detected
        #     start = process_next_batch.pop(0)
        #     end = process_next_batch.
        #     pass

        chunks = process_next_batch

        # print("proceesed this batch ",processed)
    return arranged


def sortChunks(chunks):
    """
    Sorts a collection of code "chunks" into an order that satisfies declared dependencies.

    The function implements a simple dependency resolution pass: on each iteration it selects
    chunks whose dependencies have all been "processed" and appends them to the output order.
    A chunk is considered "processed" when an assigned name can be determined from its code
    (using get_assigned_variables(chunk["code"], True)); that name (if a single string) is
    recorded and used to satisfy other chunks' dependency checks. If get_assigned_variables
    returns a set (multiple assignments) it is treated as no single assigned name (i.e. None).

    Args:
        chunks (list of dict): A list of chunk dictionaries. Each chunk is expected to contain
            at least the keys:
                - "name": a unique identifier for the chunk (not directly used for ordering
                  by the core algorithm, but typically present),
                - "code": source code string passed to get_assigned_variables(...),
                - "dependencies": a list of dependency descriptors; each dependency is expected
                  to be a mapping containing a "name" key whose value is the name of another chunk.
            Example chunk: {"name": "A", "code": "x = ...", "dependencies": [{"name": "B"}]}

    Returns:
        list of dict: A new list containing the same chunk dictionaries arranged in an order
        intended to satisfy dependencies (i.e. a chunk appears after all chunks whose names
        are listed in its "dependencies" and were successfully detected as processed).

    Side effects:
        - Prints the detected assigned name (from get_assigned_variables) for each chunk when it
          is appended to the arranged output.
        - Calls get_assigned_variables(chunk["code"], True) for each chunk to infer assignment name.

    Behavior and notes:
        - The algorithm is iterative: on each pass it collects chunks whose dependencies are
          already processed and moves them into the output. If an entire pass makes no progress,
          the remaining chunks are assumed to be in a circular dependency and are appended
          to the output as-is to break the cycle.
        - Dependency entries are expected to be mappings with a "name" key (the code checks
          dep["name"] against the processed set).
        - If a chunk's inferred assigned name is a set (multiple assignments) it is not added
          to the processed set and so will not satisfy other chunks' dependencies by name.
        - Important implementation note: during the cycle-break path the code attempts to
          add names of the remaining chunks to the processed set, but the variable used to do
          so may be incorrectly scoped/retained from an earlier loop iteration. This can cause
          incorrect entries (or missed entries) in the processed set; treat this as a potential
          bug if strict correctness on circular dependency handling is required.

    Examples:
        - Use this function when you have a collection of interdependent code chunks (e.g.
          generated code blocks) and you need a best-effort linear order that places providers
          before consumers. It does not perform full topological sorting with robust cycle
          resolution — remaining cycles are appended in arbitrary order.
    """
    processed = set()
    arranged = []


    while len(chunks) > 0:
        next_round = []
        made_progress = False

        for chunk in chunks:
            cname = get_assigned_variables(chunk["code"], True)

            if isinstance(cname,set):
                cname = None

            deps = chunk.get("dependencies", [])

            # Check whether all dependencies are processed
            all_done = all(dep["name"] in processed for dep in deps)

            if all_done:
                arranged.append(chunk)
 
                if cname:
                    processed.add(cname)
                made_progress = True
            else:
                next_round.append(chunk)

        # ---- BREAK CYCLE ----
        if not made_progress:
            # Any remaining chunks are stuck in a circular dependency.
            # Process them anyway (but only once).
            arranged.extend(next_round)
            for chunk in next_round:
                if cname:
                    processed.add(cname)
            break

        chunks = next_round

    return arranged


home = str(Path.home())
root_address = os.path.join(home,".sb_zip","speed_build_markConsultationAsPaid","lts")

files = getTemplateFileNames(root_address)

data = load_all_chunks(files)
arranged = sortChunks(data)

# # print(arranged)

# for i in arranged:
#     print(i)
#     print("\n","#"*30)


# Execute Command
import os
import sys
from pathlib import Path

from speedbuild.agents.debugCustomizer.newExec import PythonExecutor

home = str(Path.home())
express_working_dir = os.path.join(home,".sb","environment","express")#"/home/attah/.sb/environment/express/"


def executeCommand(command : str, framework:str,exec_dir:str=None):
    """
    Executes a given shell command using the PythonExecutor.

    Args:
        command (str): The shell command to execute as a string.
        framework (str) : django or express
        exec_dir (str) : where to execute command, default is root directory

    Behavior:
        - Splits the command string into a list of arguments.
        - Determines if the command should be self-exiting based on whether the last argument is "runserver".
        - Executes the command in a specified working directory and environment.
        - Return the standard output and standard error of the executed command.
    """

    home = str(Path.home())
    env = None
    wkdir = None

    if framework == "django":
        wkdir = os.path.join(home,".sb","environment","django","speedbuild_project")
        venv_path = os.path.join(home,".sb","environment","django","venv")

        if sys.platform == "win32":
            python_path =  os.path.join(venv_path, "Scripts", "python.exe")
        else:
            python_path = os.path.join(venv_path, "bin", "python")

        # Modify environment to use the venv
        env = os.environ.copy()
        env['VIRTUAL_ENV'] = venv_path

        if sys.platform == "win32":
            env['PATH'] = f"{os.path.join(venv_path, 'Scripts')};{env['PATH']}"
        else:
            env['PATH'] = f"{os.path.join(venv_path, 'bin')}:{env['PATH']}"
    elif framework == "express":
        wkdir = exec_dir if exec_dir else express_working_dir #TODO : change this to be dynamic
    else:
        return
    
    # print("executing command ",command,"\n\n\n\n")
    
    commandExecutor = PythonExecutor()
    command = command.split(" ")

    selfExiting = False if  command[-1] == "runserver" or command[-1] == "dev" else True
    

    try:
        stdout, stderr = commandExecutor.runCommand(command=command,cwd=wkdir,env=env,self_exit=selfExiting)
    except ValueError as err:
        return err


    if command[-1] == "runserver" and len(stderr) == 1:
        stderr = []

    if len(stderr) > 0:
        stderr = "\n".join(stderr)
        return f"Error : {stderr}"
    
    stdout = "\n".join(stdout)
    return f"Command was Successful and exited without error\nstdout : {stdout}"


# if __name__ == "__main__":
#     res = executeCommand("python manage.py test","django","/home/attah/.sb/environment/django/speedbuild_project")
#     print(res)


import os
import json
from pathlib import Path

from speedbuild.auth.serverCall import manageServerCall


home = str(Path.home())
template_home = os.path.join(home,".sb_zip")

def updatePushHistory(data,template_path):
    config_path = os.path.join(template_path,"history.json")
    with open(config_path,"w") as file:
        json.dump(data,file,indent=4)

def getPushHistory(template_path):
    config_path = os.path.join(template_path,"history.json")
    if not os.path.exists(config_path):
        return None
    
    with open(config_path,"r") as file:
        return json.loads(file.read())

def pushTemplate(name,framework):
    template_path = os.path.join(template_home,name)
    if not os.path.exists(template_path):
        print("Could not find template with name ",name)
        return
    
    push_history = getPushHistory(template_path)
    push_action = "update"
    
    if push_history == None:
        repo = int(input("What repo are you pushin to (enter repo id) : "))
        description = input("Enter template short description : ")
        push_history = {
            "name":name,
            "description":description,
            "framework": framework,
            "versions":[],
            "repo":repo
        }

        push_action = "create"
    else:
        repo = push_history['repo']

    filesInTemplatePath = [i for i in os.listdir(template_path) if (i.endswith(".zip") and i not in push_history['versions']) and not i.endswith("lts.zip")]

    if len(filesInTemplatePath) > 0 and "lts.zip" not in filesInTemplatePath:
        filesInTemplatePath.append("lts.zip")
    elif push_action == "create" and "lts.zip" not in filesInTemplatePath:
        filesInTemplatePath.append("lts.zip")

    filesInTemplatePath = sorted(filesInTemplatePath, key=lambda x: float(x.replace('.zip','')) if x[0].isdigit() else float('inf'))

    for zipFile in filesInTemplatePath:
        print("pushing ",zipFile)
        # send to server here
        fileFullPath = os.path.join(template_path,zipFile)

        formData = {
            "name":push_history['name'],
            "repo":push_history['repo'],
            "description":push_history['description'],
            "framework":framework,
            "version": os.path.basename(zipFile).replace(".zip","").lstrip("/"),
            "action":push_action
        }

        files = {
            "template": open(fileFullPath, "rb")
        }
        try:
            print("pushing to server")
            manageServerCall("POST","api/template/",formData,files,True)
            push_history['versions'].append(zipFile)
            suceeded = True
            if push_action == "create":
                push_action = "update"
        except ValueError as error:
            print(error)
            suceeded = False
            break
        # "name": 2,

    if suceeded:
        updatePushHistory(push_history,template_path)

# pushTemplate("test_hello","express",1)



"rough sketch"

from speedbuild.utils.parsers.python.parser import PythonBlockParser, indentCode


extra="""
from django.shortcuts import render
from django.utils import timezone
from rest_framework import status
from rest_framework.response import Response
from .models import RevenueLog
from .models import Prefrence
"""

code ="""
    if paymentStatus == True and amount >= log.amount_in_rand:
        now = timezone.now()
        date = now.date()
        month = now.date().strftime("%B")
        year = now.date().year

        log.paid = True
        log.date = date
        log.month = month
        log.year = year
        log.jannis_percentage = jannis_percentage
        log.save()

        # set consultation to paid
        consultation = log.item
        consultation.has_paid = True
        consultation.amount_paid = amount
        consultation.date_paid = timezone.now()
        consultation.save()

        # TODO : jannis revenue log here

        jannis_revenue = RevenueLog()
        jannis_revenue.for_jannis = True
        jannis_revenue.date = date
        jannis_revenue.month = month
        jannis_revenue.year = year
        jannis_revenue.revenue_for = log.revenue_for
        jannis_revenue.amount_in_rand = log.amount_in_rand
        jannis_revenue.paid = True
        jannis_revenue.item = log.item
        jannis_revenue.paid_by = log.paid_by
        jannis_revenue.jannis_percentage = jannis_percentage
        jannis_revenue.save()

        # TODO : send confirmation email
        # TODO : send zoom meeting link 
    else:
        print("Emmanuel is Good")
        num=100
"""


def getCodeChildren(code,alltheWay=False):

    """Get nested code blocks from indented code.
    This function splits a code string into chunks based on indentation levels. It can either
    preserve the original indentation or strip all whitespace from the beginning of lines.
    Args:
        code (str): The input code string to be split into chunks
        alltheWay (bool, optional): If True, strips all leading whitespace from lines.
            If False, only removes one level of indentation. Defaults to False.
    Returns:
        list: A list of code chunks, where each chunk is a string containing one or more lines
            that were at the same indentation level in the original code.
    Example:
        >>> code = '''def foo():
        ...     x = 1
        ...     if True:
        ...         y = 2
        ... '''
        >>> getCodeChildren(code)
        ['x = 1\nif True:\n    y = 2']
    """

    block = []
    chunks = []
    has_subchunk = False
    lines = code.split("\n")

    for line in lines:

        if len(line.strip()) == 0:
            block.append("\n")
            continue

        if alltheWay:
            line = line.strip()
            block.append(line)
            continue

        if line.startswith(" "):
            if not has_subchunk:
                _,line = line.split("    ",1)
                # print(line)
        else:
            has_subchunk = True
            if len(block) > 0:
                chunks.append("\n".join(block))
                block = [] 
    
        block.append(line)
    
    if len(block) > 0:
        chunks.append("\n".join(block))

    # print(len(chunks),chunks)
    return chunks

def getChildres(code):

    data_structure_open_characters = ["{","[","("]
    data_structure_close_characters = ["}","]",")"]

    code = code.strip()
    checks = ["def","class"]

    chunks = []
    skip_parsing = False

    if code.split(" ",1)[0].strip() in checks:
        segments = code.split("\n")

        if len(segments) > 1:
            chunks.append(segments[0])
            code = "\n".join(segments[1:])

            chunks = getCodeChildren(code)
 
            chunks = PythonBlockParser().parse_code("\n".join(chunks))

            if len(chunks) > 0:
                skip_parsing = True
                for i in range(0,len(chunks)):
                    chunks[i] = indentCode(chunks[i])

                chunks.insert(0,segments[0])

    elif code[-1] in data_structure_close_characters: # is a list | set | dictionary | tuple | function or class call
        opening_char = data_structure_open_characters[data_structure_close_characters.index(code[-1])]
        opening, code = code.split(opening_char,1)
        code = code[:-1]
        chunks.append(opening)
        code = getCodeChildren(code,True)
    else:
        print("here")
        code = "\n".join(getCodeChildren(code))
        chunks = getCodeChildren(code)
        skip_parsing = True

    if not skip_parsing:
        chunks = PythonBlockParser().parse_code(code)
    
    for chunk in chunks:
        print(chunk)
        print("\n","-"*30,"\n")


getChildres(code)
# print(code)


import os
import shutil



def clear_folder(folder_path):
    """
    Clear all contents of a specified folder, removing files, symlinks, and subdirectories.

    Args:
        folder_path (str): Path to the folder to be cleared.

    Returns:
        None

    Raises:
        OSError: If there are permission issues or other OS-level errors when deleting files.
        FileNotFoundError: If the specified folder path does not exist.

    Examples:
        >>> clear_folder('/path/to/folder')
        # All contents of the folder will be deleted
    """
    for file_name in os.listdir(folder_path):
        file_path = os.path.join(folder_path, file_name)
        if os.path.isfile(file_path) or os.path.islink(file_path):  
            os.remove(file_path)  # Delete files & symlinks
        elif os.path.isdir(file_path):  
            shutil.rmtree(file_path)  # Delete directories


import sys

system_prompt = f"""
You are a **smart AI agent** specialized in **Express Js** and experienced in **production-grade Express Js code modification**. 
You are part of SpeedBuild, a service that helps developers extract, customize, and reuse Express Js features using plain English commands.

## Core Mission
Extract and customize Express Js features while maintaining production reliability, performance, and Express Js best practices.

IMPORTANT:
- You must ALWAYS respond with valid JSON ONLY.
- Do not output explanations, markdown, or any text outside JSON.
- Your output will be parsed by a computer program.

Available Tools

1. read_file
   - Reads and returns the content of a file.
   - Arguments:
       file_name (str): Absolute path of the file to read.

2. update_file
   - Allows you to add, replace, or remove chunks of code in a file.
   - Requires:
       chunk_name (str): Name of the anchor chunk.
   - Behavior:
       add → write new code *before* the anchor chunk
       replace → new code *replaces* the anchor chunk
       remove → anchor chunk is *deleted*
   - The anchor chunk is mandatory for all write operations.
   - Never modify code that does not belong to a named chunk.

3. break_chunk
   - Splits an existing chunk into smaller named sub-chunks.
   - Use this when you must modify part of a chunk.
   - You may call break_chunk repeatedly until the exact code you need to modify has its own named chunk.
   - Never perform update_file operations on unnamed code.

Shell Commands:
- You CANNOT execute shell commands directly
- To request command execution, respond with EXACTLY this JSON format:
  {{
    "type": "shell_command",
    "command": "npm run dev",
    "message": "Run developtment server"
  }}
- STOP and WAIT for the execution result before continuing

Regular Response:
- for regular non command execution use the json format:
{{
    "message": <normal response>
}}

Project Files:

{files}

Note: These are the ONLY files you can modify. Always use read_file() to inspect contents before editing.

## Workflow Strategy

### 1. Understand Request
- Analyze the customization goal and scope
- Identify affected Express.js components (routes, middleware, controllers, models, services)
- Determine file processing order and dependencies

### 2. Gather Context
- Read relevant files to understand current implementation
- Check configurations, dependencies, and environment setup
- Identify potential breaking changes or dependencies

### 3. Execute Changes
- **Process dependencies first** before main feature modifications
- Follow Express Js patterns and conventions
- Add meaningful docstrings and comments for complex changes

### 4. Validate Results
- Run tests or commands to verify functionality
- Check for broken imports or syntax errors
- Ensure no regression in existing features

IMPORTANT NOTES : 
1. Always read_file() before making changes
2. Use break_chunk() to isolate the exact code you need to modify
3. Use update_file() to make the change
4. Verify your changes make syntactic and logical sense

General Rules:

- Always inspect files using read_file when needed.
- Always follow clean, production-grade NodeJs & Express Js best practices.
- Perform edits ONLY using update_file.
- Use break_chunk to isolate specific code before editing.
- If an action fails after multiple attempts:
    Respond with a JSON explaining that the action was not successful and include possible reasons.
- Never repeat failed attempts unnecessarily.

---
NOTE : The application is running on"""+ "Windows" if sys.platform == "win32" else "Unix-based os"+ """!!!
---

"""


import os
import shutil
import sys
import asyncio

from pathlib import Path
from rich.prompt import Prompt

from speedbuild.agents.config import getLLMConfig, setSpeedbuildConfig
from speedbuild.utils.cleanup.template.clear_folder import clear_folder

from .frameworks.express.deployment.deployJS import deployExpressFeature
from .frameworks.django.extraction.extract_features import createTemplate
from .frameworks.express.extraction.extractjs import startExpressExtraction
from .frameworks.django.deployment.debug_and_customize import testCustomizeAndDeployDjangoFeature 

from .utils.undo.undo import undoSBDeploy
from .utils.repo_templates.push import pushTemplate
from .utils.package_utils import getPackageNameMapping
from .utils.repo_templates.pull import pullTemplateFromHub

from .auth.sb_user import sbAuth
from .auth.auth import manageAuth
from .auth.cli_utils import get_config
from .auth.serverCall import manageServerCall, pingServer


def logUserCommand(args):
    command = "speedbuild " + " ".join(args[1:])
    manageServerCall("POST","api/log-action/",{"command":command},{},True)

def extractFeature(project_root,args,append_root=False):
    try:
        target = args[2]
        extract_from = args[3]
        framework = args[4]

        framework = framework.replace("--","").lower()

        # clear extraction output folder here
        output_folder = os.path.join(project_root,"output")
        if os.path.exists(output_folder):
            clear_folder(output_folder)
            shutil.rmtree(output_folder)

        # set run_documentation env to its default
        os.environ['run_documentation'] = "False"

        if framework == "django":
            # Extract django feature
            # logger.start_status("Extracting django feature")
            try:
                project_name = os.path.basename(project_root)
                packageToNameMapping,installedPackages = getPackageNameMapping(project_root)
                createTemplate(extract_from, target,project_name,packageToNameMapping,project_root,append_root,installedPackages)
            except ValueError as e:
                print(e)
            
        elif framework == "express":
            #Extract Express feature
            asyncio.run(startExpressExtraction(target,extract_from,project_root))

        # if framework in ['django','express']:
        #     logUserCommand(args)

    except IndexError:
        print("Usage : python final.py <what_to_extract> <extraction_entry_point> --framework")


def deployFeature(project_root,args):
    try:
        template = args[2]
        framework = args[3]
        framework = framework.replace("--","").lower()
        print(f"Deploying {template}")
        
        customization_request = Prompt.ask("[cyan bold] Will you like to customize this feature (yes/no) [/cyan bold]")
        while customization_request.lower().strip() not in ['yes','no']:
            customization_request = Prompt.ask("[cyan] Please enter a valid response (yes/no)[/cyan]")

        customization_request = customization_request.strip() == "yes"

        if framework == "django":
            # implementFeature(template)
            app_name = Prompt.ask("[cyan bold] Which django app do you want to implement feature in [/cyan bold] ")
            app_path = os.path.join(project_root,app_name)
            if os.path.exists(app_path) and os.path.isdir(app_path):
                testCustomizeAndDeployDjangoFeature(template,project_root,app_name,customization_request)
            else:
                print("Please Enter a valid app name")
        elif framework == "express":
            asyncio.run(deployExpressFeature(template,project_root,customization_request)) #true is so we customize otherwise use false

        # if framework in ['django','express']:
        #     logUserCommand(args)

    except IndexError:
        print("Usage : python final.py deploy <template_name>")

def listExtractedFeature(args):
    print("Listing All Extracted Templates")
    user_home = str(Path.home())
    templates_dir = f"{user_home}/.sb_zip"

    if not os.path.exists(templates_dir):
        os.makedirs(templates_dir)

    count = 1
    templates = os.listdir(templates_dir)

    for template in templates:
        # if template.startswith("speed_build"):
        #     print(f"    {count}) {template}")
        #     count += 1
        print(f"    {count}) {template}")
        count += 1
        
    print("\n")

def start():
    args = sys.argv
    setSpeedbuildConfig()
    getLLMConfig()
    current_path = os.path.abspath(".")


    try:
        # TODO : remove this
        api_key = os.environ.get("OPENAI_API_KEY",None)
        if api_key == None:
            config = get_config()
            if "openai" in config['llm_keys'].keys():
                os.environ["OPENAI_API_KEY"] = config["llm_keys"]['openai']
    except:
        pass

    try:
        command = args[1]

        if command == "extract":
            # sb.py extract / views.py --express
            extractFeature(current_path,args)

        elif command == "deploy":
            # sb.py deploy template_name
            deployFeature(current_path,args)

        elif command == "list":
            listExtractedFeature(args)

        elif command == "undo":
            print("performing undo")
            undoSBDeploy()

        elif command == "auth":
            asyncio.run(sbAuth())

        elif command == "ping":
            pingServer()
            
        elif command == "setup":
            print("Seting up speedbuild")
            asyncio.run(manageAuth("register"))

        elif command == "push":
            try:
                template = args[2]
                framework = args[3].replace("-","").strip()

                if framework not in ["express","django"]:
                    print("Please enter a valid framework")
                    return
                
                user_home = str(Path.home())
                template_dir = os.path.join(user_home,".sb_zip",template)

                if not os.path.exists(template_dir):
                    print(f"Invalid template : {template}")
                    return
                
                pushTemplate(template,framework)
                
            except IndexError:
                print("Usage : speedbuild push <template_name> --framework")

        elif command == "pull":
            try:
                template = args[2]
                if "/" not in template:
                    print("Usage : speedbuild pull <template_repo>/<template_name>")
                    return
                
                pullTemplateFromHub(template)
            except IndexError:
                print("Usage : speedbuild pull <template_repo>/<template_name>")
            except ValueError as err:
                print("Pull error :",err)

        elif command == "use":
            try:
                template = args[2]
                framework = args[3]
                framework = framework.replace("--","").lower()

                if "/" not in template:
                    print("Usage : speedbuild use <template_repo>/<template_name> --framework")
                    return
        
                template_name = pullTemplateFromHub(template)
                print("pulled template ", template_name)
                deployFeature(current_path,["speedbuild","deploy",template_name, framework])

            except IndexError:
                print("Usage : speedbuild use <template_repo>/<template_name> --framework")

            except ValueError as err:
                print("Pull error :",err)


    except IndexError as e:
        print("Available commands :\n- extract\n- deploy\n- list\n- undo\n- auth\n- ping\n- setup\n- push\n- pull")

# start() # remove this


import os
import yaml
import json
import shutil
from pathlib import Path 

from speedbuild.utils.markdown.yaml import read_yaml 
from speedbuild.utils.cli.output import StatusManager
from speedbuild.utils.cleanup.template.clear_folder import clear_folder 
from speedbuild.frameworks.django.utils.venv_utils import get_activated_venv 
from speedbuild.frameworks.django.utils.django_utils import ManageDjangoSettings 
from speedbuild.utils.template.file import findFilePath, get_template_output_path 
from speedbuild.utils.template.template_update import checkTemplateForVersionUpdate 
from speedbuild.frameworks.django.extraction.feature_dependencies import arrangeChunks 
from speedbuild.frameworks.django.extraction.extract import extract_feature_code_and_dependencies, getTemplateFileNames, getURLForFeature 

from .test_agent import SpeedBuildTestAgent


def debugTemplate(template,packageToNameMapping,installedPackages):
    debugger = SpeedBuildTestAgent()
    full_template_path = os.path.join(debugger.home,".sb_zip",template,"lts.zip")

    if not os.path.exists(full_template_path):
        print(f"Template with name {template} does not exist")
        return
    
    ran_debug, packages_before_debug = debugger.TestFeature(template)

    if ran_debug: # if AI debugged code, update extracted feature else do nothing
        packages_after_debug = debugger.getPackagesinVenv()
        new_packages = list(packages_after_debug.difference(packages_before_debug)) #packages installed by debug
        if len(new_packages) > 0:
            # save new packages
            save_path = os.path.join(debugger.home,".sb","django_package_cache.json")
            if not(os.path.dirname(save_path)):
                os.makedirs(os.path.dirname(save_path))

            with open(save_path,"w") as file:
                json.dump({"packages":new_packages},file,indent=4)

        path_to_template = os.path.join(debugger.home,".sb","sb_extracted")
        template_root = os.path.join(path_to_template,template)
        feature_name = template.replace("speed_build_","")
        config_file_path = os.path.join(template_root,f"sb_{feature_name}.yaml")
        data = read_yaml(config_file_path)

        feature_path = os.path.join("speedbuild_app",os.path.basename(data['feature_file_path']))
        project_root = os.path.join(debugger.home,".sb","environment","django","speedbuild_project")

        # delete old zip file before repackaging
        os.remove(full_template_path)

        # re-extract debug feature
        createTemplate(
            feature_path, 
            feature_name,
            "speedbuild_project",
            packageToNameMapping,
            project_root,
            True,
            installedPackages,
            run_debug=False
        )
    else:
        logger = StatusManager()
        logger.print_message(f"Template `{template}` Created")

# remove comments from code before splitting blocks
def create_temp_from_feature(
        project_path,
        project_name,
        feature_name,
        feature_file_path,
        template_django_apps,
        installedPackages,
        skip_debug
    ):

    venv = get_activated_venv()
    
    # get settings conf from django.conf.settings TODO
    app_name = os.path.dirname(feature_file_path)

    if app_name == os.path.abspath(app_name): #incase app_name holds an abs path just get the base name
        app_name = os.path.basename(app_name)


    findFilePath(project_path, "urls.py")

    # use conditional statement to check if file is found
    project_name = findFilePath(project_path, "asgi.py")

    if len(project_name) == 0:
        raise(ValueError("Cannot find django project settings file"))
    
    template_dep = set()
    installed_apps = set()
    project_name = project_name[0].split("/")[0]
    settings_path = os.path.join(project_path, project_name,"settings.py")

    project_dir = [folder for folder in os.listdir(project_path) if os.path.isdir(os.path.join(project_path,folder))]

    settings = ManageDjangoSettings(settings_path,venv)

    processed = set()


    # start; make recursive

    output_folder_name = "sb_output_"+feature_name

    template_settings_import,template_confs,template_dep = extract_feature_code_and_dependencies(
        feature_file_path,
        feature_name,
        project_path,
        project_dir,
        app_name,
        template_django_apps,
        settings,
        project_name,
        installed_apps,
        template_dep,
        processed,
        [],
        None,
        output_folder_name,
        installedPackages
    )

    # end recursion
    output_dir = os.path.join(".","output",output_folder_name)

    # Create template yaml file
    template_confs = arrangeChunks(template_confs,[],[])

    django_package_cache_path = os.path.join(str(Path.home()),".sb","django_package_cache.json")

    if os.path.exists(django_package_cache_path):
        with open(django_package_cache_path) as package_cache:
            package_cache = json.loads(package_cache.read())
            template_dep = template_dep.union(set(package_cache['packages']))

        # remove cache packages
        os.remove(django_package_cache_path)
    
    data = {
        "feature" : feature_name,
        "feature_file_path" : feature_file_path.replace(project_path,""),
        "source_project":project_name,
        "dependencies" : list(template_dep),
        "settings" : {
            "imports" : list(sorted(set(template_settings_import))),
            "installed_apps" : list(installed_apps),
            "middlewares" : settings.getTemplateMiddleWare(list(installed_apps)),
            "configurations" : template_confs#list(set(template_confs))
        }
    }

    yaml_file_path = os.path.join(output_dir,f"sb_{feature_name}.yaml")

    with open(yaml_file_path, 'w') as file:
        yaml.dump(data, file, default_flow_style=False,sort_keys=False)


    if os.path.exists(os.path.join(output_dir,"sb_app","views.py")):
        # if views.py in models, generate url for feature
        getURLForFeature(feature_name,feature_file_path,app_name,project_path,output_folder_name)

    # create template documentations here
    tem_files = getTemplateFileNames(output_dir)
    tem_files = sorted(tem_files, key=lambda x: not x.startswith('.sb_utils')) #process dependencies first
    # Exclude the yaml file

    # for file in tem_files:
    #     file_name = file.split("/")
    #     main_file = file_name.pop().split(".")[0] #remove extention from file name
    #     file_name.append(main_file)
    #     file_name = "_".join(file_name)
  
    #     main_dir = "./output/documentation"
    #     if not os.path.exists(main_dir):
    #         os.makedirs(main_dir)

    #     # write to file  TODO : leave commented for now
    #     with open(f"{main_dir}/{file_name}.md","w") as new_file:
    #         # get and write doc to file
    #         pass
    #         # doc = documentationAgent(file)
    #         # new_file.write(doc)

    # Save template to user main dir in the .sb_zip folder
    user_dir = str(Path.home())
    user_dir = os.path.join(user_dir,".sb_zip")

    if not os.path.exists(user_dir):
        os.makedirs(user_dir)

    template_path = get_template_output_path(feature_name)

    shutil.make_archive(template_path, 'zip', output_dir)


    # delete output folder
    clear_folder(output_dir)
    
    # structure templates to include versions
    if skip_debug:
        checkTemplateForVersionUpdate(template_path)
    else:
        shutil.move(f"{template_path}.zip", os.path.join(template_path, "lts.zip"))

    return template_path


def createTemplate(
        file_path, 
        feature,
        project_name,
        packageToNameMapping,
        project_root,
        append_root=False,
        installedPackages={},
        run_debug=True
    ):

    logger = StatusManager()

    if file_path.endswith("/"):
        file_path = file_path.lstrip("/")

    if append_root:
        file_path = os.path.join(project_root,file_path)
    else:
        file_path = file_path.replace(project_root,"")

    if run_debug:
        logger.start_status("Extracting Feature")
    else:
        logger.start_status("Re-extracting debugged feature")

    # print("\n##### Extracting Feature : ",feature," ##### from ",file_path)

    if not os.path.exists(file_path):
        print("invalid feature")
        return

    template_path = create_temp_from_feature(
        project_root,
        project_name,
        feature,
        file_path,
        packageToNameMapping,
        installedPackages,
        run_debug
    )

    logger.stop_status("Feature Extracted")


    # debug template to catch error our algorithm missed
    if run_debug:
        os.environ['run_documentation'] = "True"
        debugTemplate(os.path.basename(template_path),packageToNameMapping,installedPackages)

    if not run_debug:
        os.environ['run_documentation'] = "False"
        # TODO : run index for RAG
        logger.print_message(f"Template `{template_path}` Created")



import json
import os
import requests
from pathlib import Path

from speedbuild.auth.serverCall import getRequestHeaders, manageServerCall,url_address


"""
    Code to pull template from speedbuild hub. 
"""

def download_file(url, save_folder, filename,data):
    data['download'] = True

    # Ensure the folder exists
    os.makedirs(save_folder, exist_ok=True)
    
    # If no filename provided, extract from URL
    if not filename:
        filename = url.split("/")[-1] or "downloaded_file"
    
    file_path = os.path.join(save_folder, filename)

    # get access_token

    # Stream download to handle large files
    with requests.post(url,data=data,headers=getRequestHeaders(), stream=True) as response:
        response.raise_for_status()  # Raise error for bad responses (4xx, 5xx)
        with open(file_path, "wb") as file:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:  # Filter out keep-alive chunks
                    file.write(chunk)

    print(f"✅ File saved to: {file_path}")
    return file_path

def pullTemplateFromHub(template_name):
    """
        repo_name
        template_name
        version
    """

    data = {
        "repo_name" : os.path.dirname(template_name),
    }

    template_name = os.path.basename(template_name)
    version = "lts"

    if ":" in template_name:
        template_name,version = template_name.split(":")
        data['version'] = version

    data["template_name"] = os.path.basename(template_name)

    if not template_name.startswith("speed_build_"):
        template_name = "speed_build_"+template_name

    save_path = os.path.join(str(Path.home()),".sb_zip",template_name)

    # TODO get and save template history
    success, res = manageServerCall("POST","api/pull-template/",data,{},True)

    if success:
        # save history file
        if not os.path.exists(save_path):
            os.makedirs(save_path,exist_ok=True)

        with open(os.path.join(save_path,"history.json"),"w") as file:
            json.dump(res,file,indent=4)
    else:
        raise ValueError(res)

    url = url_address + "api/pull-template/"
    
    # download template
    save_full_path = os.path.join(save_path,f"{version}.zip")
    save_folder = os.path.dirname(save_full_path)
    filename = os.path.basename(save_full_path)

    download_file(url,save_folder,filename,data)

    return template_name




from typing import Optional
from pydantic import BaseModel, Field

from langchain.chat_models import init_chat_model
from langchain_core.messages import SystemMessage, HumanMessage
from speedbuild.agents.config import getLLMConfig

system_prompt = '''
You are **DocuSmith**, an AI agent that generates documentation for Python (Django) and JavaScript/TypeScript (Express) code.

Your job:

1. **Decide if the code requires documentation.**

   * If the code is trivial (one-liners, simple assignments, imports, or obvious helpers), respond with:

     * `has_documentation = false`
     * `documentation = null`

2. **Document only meaningful code** such as functions, classes, methods,
   Django views/models/serializers, Express routes/controllers/middleware, or any code with real logic.

3. **Generate documentation in the correct format:**

**Python (Django) → Google-style docstring**

    """
    Summary.

    Args:
        param (type): description
    Returns:
        type: description
    """

**JS/TS (Express) → JSDoc**

    /**
    * Summary.
    * @param {...} ...
    * @returns {...}
    */

4. **Do not modify the original code.**
   Only return documentation text.

5. **Do not hallucinate behavior** not present in the code.

6. **Your response MUST strictly be a valid `DocumentationOutput` object:**

    ```json
    {
        "has_documentation": true or false,
        "documentation": "string or null"
    }
    ```

'''

class DocumentationOutput(BaseModel):
    has_documentation : bool = Field(description="determine if given code should be documented")
    documentation : Optional[str] = Field(description="code documentation")

def DocumentCode(code):
    model_provider, model = getLLMConfig()    
    model = init_chat_model(model_provider=model_provider, model=model)
    model = model.with_structured_output(DocumentationOutput)

    return model.invoke([
        SystemMessage(content=system_prompt),
        HumanMessage(content=code)
    ])
